from collections import defaultdict
from pprint import pprint
from typing import Dict, List
from dataclasses import dataclass

import numpy as np
import torch
from tqdm import tqdm
from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer,
    AutoConfig,
    T5ForConditionalGeneration,
)

from a2t.base import Classifier, np_softmax


@dataclass
class REInputFeatures:
    subj: str
    obj: str
    context: str
    pair_type: str = None
    label: str = None


class _NLIRelationClassifier(Classifier):
    def __init__(
        self,
        labels: List[str],
        *args,
        pretrained_model: str = "roberta-large-mnli",
        use_cuda=True,
        half=False,
        verbose=True,
        negative_threshold=0.95,
        negative_idx=0,
        max_activations=np.inf,
        valid_conditions=None,
        **kwargs,
    ):
        super().__init__(
            labels,
            pretrained_model=pretrained_model,
            use_cuda=use_cuda,
            verbose=verbose,
            half=half,
        )
        # self.ent_pos = entailment_position
        # self.cont_pos = -1 if self.ent_pos == 0 else 0
        self.negative_threshold = negative_threshold
        self.negative_idx = negative_idx
        self.max_activations = max_activations
        self.n_rel = len(labels)
        # for label in labels:
        #     assert '{subj}' in label and '{obj}' in label

        if valid_conditions:
            self.valid_conditions = {}
            rel2id = {r: i for i, r in enumerate(labels)}
            self.n_rel = len(rel2id)
            for relation, conditions in valid_conditions.items():
                if relation not in rel2id:
                    continue
                for condition in conditions:
                    if condition not in self.valid_conditions:
                        self.valid_conditions[condition] = np.zeros(self.n_rel)
                        self.valid_conditions[condition][rel2id["no_relation"]] = 1.0
                    self.valid_conditions[condition][rel2id[relation]] = 1.0

        else:
            self.valid_conditions = None

        def idx2label(idx):
            return self.labels[idx]

        self.idx2label = np.vectorize(idx2label)

    def _initialize(self, pretrained_model):
        self.tokenizer = AutoTokenizer.from_pretrained(pretrained_model)
        self.model = AutoModelForSequenceClassification.from_pretrained(pretrained_model)
        self.config = AutoConfig.from_pretrained(pretrained_model)
        self.ent_pos = self.config.label2id.get("ENTAILMENT", self.config.label2id.get("entailment", None))
        if self.ent_pos is None:
            raise ValueError("The model config must contain ENTAILMENT label in the label2id dict.")
        else:
            self.ent_pos = int(self.ent_pos)

    def _run_batch(self, batch, multiclass=False):
        with torch.no_grad():
            input_ids = self.tokenizer.batch_encode_plus(batch, padding=True, truncation=True)
            input_ids = torch.tensor(input_ids["input_ids"]).to(self.device)
            output = self.model(input_ids)[0].detach().cpu().numpy()
            if multiclass:
                output = np.exp(output) / np.exp(output).sum(
                    -1, keepdims=True
                )  # np.exp(output[..., [self.cont_pos, self.ent_pos]]).sum(-1, keepdims=True)
            output = output[..., self.ent_pos].reshape(input_ids.shape[0] // len(self.labels), -1)

        return output

    def __call__(
        self,
        features: List[REInputFeatures],
        batch_size: int = 1,
        multiclass=False,
    ):
        if not isinstance(features, list):
            features = [features]

        batch, outputs = [], []
        for i, feature in tqdm(enumerate(features), total=len(features)):
            sentences = [
                f"{feature.context} {self.tokenizer.sep_token} {label_template.format(subj=feature.subj, obj=feature.obj)}."
                for label_template in self.labels
            ]
            batch.extend(sentences)

            if (i + 1) % batch_size == 0:
                output = self._run_batch(batch, multiclass=multiclass)
                outputs.append(output)
                batch = []

        if len(batch) > 0:
            output = self._run_batch(batch, multiclass=multiclass)
            outputs.append(output)

        outputs = np.vstack(outputs)

        return outputs

    def _apply_negative_threshold(self, probs):
        activations = (probs >= self.negative_threshold).sum(-1).astype(np.int)
        idx = np.logical_or(
            activations == 0, activations >= self.max_activations
        )  # If there are no activations then is a negative example, if there are too many, then is a noisy example
        probs[idx, self.negative_idx] = 1.00
        return probs

    def _apply_valid_conditions(self, probs, features: List[REInputFeatures]):
        mask_matrix = np.stack(
            [self.valid_conditions.get(feature.pair_type, np.zeros(self.n_rel)) for feature in features],
            axis=0,
        )
        probs = probs * mask_matrix

        return probs

    def predict(
        self,
        contexts: List[str],
        batch_size: int = 1,
        return_labels: bool = True,
        return_confidences: bool = False,
        topk: int = 1,
    ):
        output = self(contexts, batch_size)
        topics = np.argsort(output, -1)[:, ::-1][:, :topk]
        if return_labels:
            topics = self.idx2label(topics)
        if return_confidences:
            topics = np.stack((topics, np.sort(output, -1)[:, ::-1][:, :topk]), -1).tolist()
            topics = [
                [(int(label), float(conf)) if not return_labels else (label, float(conf)) for label, conf in row]
                for row in topics
            ]
        else:
            topics = topics.tolist()
        if topk == 1:
            topics = [row[0] for row in topics]

        return topics


class _GenerativeNLIRelationClassifier(_NLIRelationClassifier):
    """_GenerativeNLIRelationClassifier

    This class is intended to be use with T5 like NLI models.

    TODO: Test
    """

    def _initialize(self, pretrained_model):

        if "t5" not in pretrained_model:
            raise NotImplementedError(
                f"This implementation is not available for {pretrained_model} yet. "
                "Use a t5-[small-base-large] model instead."
            )

        self.tokenizer = AutoTokenizer.from_pretrained(pretrained_model)
        self.model = T5ForConditionalGeneration.from_pretrained(pretrained_model)
        self.config = AutoConfig.from_pretrained(pretrained_model)
        self._get_entailment_neutral_contradiction_token_id()

    def _run_batch(self, batch, multiclass=False):
        with torch.no_grad():
            input_ids = self.tokenizer.batch_encode_plus(batch, padding=True, truncation=True)
            input_ids = torch.tensor(input_ids["input_ids"]).to(self.device)
            decoder_input_ids = self.tokenizer.batch_encode_plus(len(batch) * ["<pad>"], padding=True, truncation=True)
            decoder_input_ids = torch.tensor(decoder_input_ids["input_ids"]).to(self.device)
            output = self.model(input_ids, decoder_input_ids=decoder_input_ids)[0].detach().cpu().numpy()
            output = self._vocab_to_class_logits(output)
            if multiclass:
                output = np.exp(output) / np.exp(output).sum(
                    -1, keepdims=True
                )  # np.exp(output[..., [self.cont_pos, self.ent_pos]]).sum(-1, keepdims=True)
            output = output[..., 0].reshape(input_ids.shape[0] // len(self.labels), -1)

        return output

    def _get_entailment_neutral_contradiction_token_id(self):
        class_ids = self.tokenizer(["entailment", "neutral", "contradiction"]).input_ids
        (
            self.entailment_token_id,
            self.neutral_token_id,
            self.contradiction_token_id,
        ) = [ids[0] for ids in class_ids]
        assert (
            (self.entailment_token_id != self.neutral_token_id)
            and (self.entailment_token_id != self.contradiction_token_id)
            and (self.neutral_token_id != self.contradiction_token_id)
        )

    def _vocab_to_class_logits(self, outputs):
        class_logits = outputs[
            :,
            0,
            [
                self.entailment_token_id,
                self.neutral_token_id,
                self.contradiction_token_id,
            ],
        ]
        return class_logits


class NLIRelationClassifier(_NLIRelationClassifier):
    def __init__(
        self,
        labels: List[str],
        *args,
        pretrained_model: str = "roberta-large-mnli",
        **kwargs,
    ):
        super().__init__(labels, *args, pretrained_model=pretrained_model, **kwargs)

    def __call__(
        self,
        features: List[REInputFeatures],
        batch_size: int = 1,
        multiclass=True,
    ):
        outputs = super().__call__(features=features, batch_size=batch_size, multiclass=multiclass)
        outputs = np_softmax(outputs) if not multiclass else outputs
        outputs = self._apply_negative_threshold(outputs)

        return outputs


class NLIRelationClassifierWithMappingHead(_NLIRelationClassifier):
    def __init__(
        self,
        labels: List[str],
        template_mapping: Dict[str, str],
        pretrained_model: str = "roberta-large-mnli",
        valid_conditions: Dict[str, list] = None,
        *args,
        **kwargs,
    ):

        self.template_mapping_reverse = defaultdict(list)
        for key, value in template_mapping.items():
            for v in value:
                self.template_mapping_reverse[v].append(key)
        self.new_topics = list(self.template_mapping_reverse.keys())

        self.target_labels = labels
        self.new_labels2id = {t: i for i, t in enumerate(self.new_topics)}
        self.mapping = defaultdict(list)
        for key, value in template_mapping.items():
            self.mapping[key].extend([self.new_labels2id[v] for v in value])

        super().__init__(
            self.new_topics,
            *args,
            pretrained_model=pretrained_model,
            valid_conditions=None,
            **kwargs,
        )

        if valid_conditions:
            self.valid_conditions = {}
            rel2id = {r: i for i, r in enumerate(labels)}
            self.n_rel = len(rel2id)
            for relation, conditions in valid_conditions.items():
                if relation not in rel2id:
                    continue
                for condition in conditions:
                    if condition not in self.valid_conditions:
                        self.valid_conditions[condition] = np.zeros(self.n_rel)
                        self.valid_conditions[condition][rel2id["no_relation"]] = 1.0
                    self.valid_conditions[condition][rel2id[relation]] = 1.0

        else:
            self.valid_conditions = None

        def idx2label(idx):
            return self.target_labels[idx]

        self.idx2label = np.vectorize(idx2label)

    def __call__(self, features: List[REInputFeatures], batch_size=1, multiclass=True):
        outputs = super().__call__(features, batch_size, multiclass)
        outputs = np.hstack(
            [
                np.max(outputs[:, self.mapping[label]], axis=-1, keepdims=True)
                if label in self.mapping
                else np.zeros((outputs.shape[0], 1))
                for label in self.target_labels
            ]
        )
        outputs = np_softmax(outputs) if not multiclass else outputs

        if self.valid_conditions:
            outputs = self._apply_valid_conditions(outputs, features)

        outputs = self._apply_negative_threshold(outputs)

        return outputs


class GenerativeNLIRelationClassifier(_GenerativeNLIRelationClassifier, NLIRelationClassifier):
    pass


class GenerativeNLIRelationClassifierWithMappingHead(_GenerativeNLIRelationClassifier, NLIRelationClassifierWithMappingHead):
    pass


if __name__ == "__main__":

    labels = [
        "no_relation",
        "org:alternate_names",
        "org:city_of_headquarters",
        "org:country_of_headquarters",
        "org:dissolved",
        "org:founded",
        "org:founded_by",
        "org:member_of",
        "org:members",
        "org:number_of_employees/members",
        "org:parents",
        "org:political/religious_affiliation",
        "org:shareholders",
        "org:stateorprovince_of_headquarters",
        "org:subsidiaries",
        "org:top_members/employees",
        "org:website",
        "per:age",
        "per:alternate_names",
        "per:cause_of_death",
        "per:charges",
        "per:children",
        "per:cities_of_residence",
        "per:city_of_birth",
        "per:city_of_death",
        "per:countries_of_residence",
        "per:country_of_birth",
        "per:country_of_death",
        "per:date_of_birth",
        "per:date_of_death",
        "per:employee_of",
        "per:origin",
        "per:other_family",
        "per:parents",
        "per:religion",
        "per:schools_attended",
        "per:siblings",
        "per:spouse",
        "per:stateorprovince_of_birth",
        "per:stateorprovince_of_death",
        "per:stateorprovinces_of_residence",
        "per:title",
    ]
    template_mapping = {
        "per:alternate_names": ["{subj} is also known as {obj}"],
        "per:date_of_birth": [
            "{subj}’s birthday is on {obj}",
            "{subj} was born in {obj}",
        ],
        "per:age": ["{subj} is {obj} years old"],
        "per:country_of_birth": ["{subj} was born in {obj}"],
        "per:stateorprovince_of_birth": ["{subj} was born in {obj}"],
        "per:city_of_birth": ["{subj} was born in {obj}"],
        "per:origin": ["{obj} is the nationality of {subj}"],
        "per:date_of_death": ["{subj} died in {obj}"],
        "per:country_of_death": ["{subj} died in {obj}"],
        "per:stateorprovince_of_death": ["{subj} died in {obj}"],
        "per:city_of_death": ["{subj} died in {obj}"],
        "per:cause_of_death": ["{obj} is the cause of {subj}’s death"],
        "per:countries_of_residence": [
            "{subj} lives in {obj}",
            "{subj} has a legal order to stay in {obj}",
        ],
        "per:stateorprovinces_of_residence": [
            "{subj} lives in {obj}",
            "{subj} has a legal order to stay in {obj}",
        ],
        "per:cities_of_residence": [
            "{subj} lives in {obj}",
            "{subj} has a legal order to stay in {obj}",
        ],
        "per:schools_attended": [
            "{subj} studied in {obj}",
            "{subj} graduated from {obj}",
        ],
        "per:title": ["{subj} is a {obj}"],
        "per:employee_of": [
            "{subj} is member of {obj}",
            "{subj} is an employee of {obj}",
        ],
        "per:religion": [
            "{subj} belongs to {obj} religion",
            "{obj} is the religion of {subj}",
            "{subj} believe in {obj}",
        ],
        "per:spouse": [
            "{subj} is the spouse of {obj}",
            "{subj} is the wife of {obj}",
            "{subj} is the husband of {obj}",
        ],
        "per:children": [
            "{subj} is the parent of {obj}",
            "{subj} is the mother of {obj}",
            "{subj} is the father of {obj}",
            "{subj} is the son of {obj}",
            "{subj} is the daughter of {obj}",
        ],
        "per:siblings": [
            "{subj} and {obj} are siblings",
            "{subj} is brother of {obj}",
            "{subj} is sister of {obj}",
        ],
        "per:other_family": [
            "{subj} and {obj} are family",
            "{subj} is a brother in law of {obj}",
            "{subj} is a sister in law of {obj}",
            "{subj} is the cousin of {obj}",
            "{subj} is the uncle of {obj}",
            "{subj} is the aunt of {obj}",
            "{subj} is the grandparent of {obj}",
            "{subj} is the grandmother of {obj}",
            "{subj} is the grandson of {obj}",
            "{subj} is the granddaughter of {obj}",
        ],
        "per:charges": [
            "{subj} was convicted of {obj}",
            "{obj} are the charges of {subj}",
        ],
        "org:alternate_names": ["{subj} is also known as {obj}"],
        "org:political/religious_affiliation": [
            "{subj} has political affiliation with {obj}",
            "{subj} has religious affiliation with {obj}",
        ],
        "org:top_members/employees": [
            "{obj} is a high level member of {subj}",
            "{obj} is chairman of {subj}",
            "{obj} is president of {subj}",
            "{obj} is director of {subj}",
        ],
        "org:number_of_employees/members": [
            "{subj} employs nearly {obj} people",
            "{subj} has about {obj} employees",
        ],
        "org:members": ["{obj} is member of {subj}", "{obj} joined {subj}"],
        "org:member_of": ["{subj} is member of {obj}", "{subj} joined {obj}"],
        "org:subsidiaries": [
            "{obj} is a subsidiary of {subj}",
            "{obj} is a branch of {subj}",
        ],
        "org:parents": [
            "{subj} is a subsidiary of {obj}",
            "{subj} is a branch of {obj}",
        ],
        "org:founded_by": [
            "{subj} was founded by {obj}",
            "{obj} founded {subj}",
        ],
        "org:founded": [
            "{subj} was founded in {obj}",
            "{subj} was formed in {obj}",
        ],
        "org:dissolved": [
            "{subj} existed until {obj}",
            "{subj} disbanded in {obj}",
            "{subj} dissolved in {obj}",
        ],
        "org:country_of_headquarters": [
            "{subj} has its headquarters in {obj}",
            "{subj} is located in {obj}",
        ],
        "org:stateorprovince_of_headquarters": [
            "{subj} has its headquarters in {obj}",
            "{subj} is located in {obj}",
        ],
        "org:city_of_headquarters": [
            "{subj} has its headquarters in {obj}",
            "{subj} is located in {obj}",
        ],
        "org:shareholders": ["{obj} holds shares in {subj}"],
        "org:website": [
            "{obj} is the URL of {subj}",
            "{obj} is the website of {subj}",
        ],
    }

    rules = {
        "per:alternate_names": ["PERSON:PERSON", "PERSON:MISC"],
        "per:date_of_birth": ["PERSON:DATE"],
        "per:age": ["PERSON:NUMBER", "PERSON:DURATION"],
        "per:country_of_birth": ["PERSON:COUNTRY"],
        "per:stateorprovince_of_birth": ["PERSON:STATE_OR_PROVINCE"],
        "per:city_of_birth": ["PERSON:CITY"],
        "per:origin": [
            "PERSON:NATIONALITY",
            "PERSON:COUNTRY",
            "PERSON:LOCATION",
        ],
        "per:date_of_death": ["PERSON:DATE"],
        "per:country_of_death": ["PERSON:COUNTRY"],
        "per:stateorprovince_of_death": ["PERSON:STATE_OR_PROVICE"],
        "per:city_of_death": ["PERSON:CITY", "PERSON:LOCATION"],
        "per:cause_of_death": ["PERSON:CAUSE_OF_DEATH"],
        "per:countries_of_residence": ["PERSON:COUNTRY", "PERSON:NATIONALITY"],
        "per:stateorprovinces_of_residence": ["PERSON:STATE_OR_PROVINCE"],
        "per:cities_of_residence": ["PERSON:CITY", "PERSON:LOCATION"],
        "per:schools_attended": ["PERSON:ORGANIZATION"],
        "per:title": ["PERSON:TITLE"],
        "per:employee_of": ["PERSON:ORGANIZATION"],
        "per:religion": ["PERSON:RELIGION"],
        "per:spouse": ["PERSON:PERSON"],
        "per:children": ["PERSON:PERSON"],
        "per:siblings": ["PERSON:PERSON"],
        "per:other_family": ["PERSON:PERSON"],
        "per:charges": ["PERSON:CRIMINAL_CHARGE"],
        "org:alternate_names": [
            "ORGANIZATION:ORGANIZATION",
            "ORGANIZATION:MISC",
        ],
        "org:political/religious_affiliation": [
            "ORGANIZATION:RELIGION",
            "ORGANIZATION:IDEOLOGY",
        ],
        "org:top_members/employees": ["ORGANIZATION:PERSON"],
        "org:number_of_employees/members": ["ORGANIZATION:NUMBER"],
        "org:members": ["ORGANIZATION:ORGANIZATION", "ORGANIZATION:COUNTRY"],
        "org:member_of": [
            "ORGANIZATION:ORGANIZATION",
            "ORGANIZATION:COUNTRY",
            "ORGANIZATION:LOCATION",
            "ORGANIZATION:STATE_OR_PROVINCE",
        ],
        "org:subsidiaries": [
            "ORGANIZATION:ORGANIZATION",
            "ORGANIZATION:LOCATION",
        ],
        "org:parents": ["ORGANIZATION:ORGANIZATION", "ORGANIZATION:COUNTRY"],
        "org:founded_by": ["ORGANIZATION:PERSON"],
        "org:founded": ["ORGANIZATION:DATE"],
        "org:dissolved": ["ORGANIZATION:DATE"],
        "org:country_of_headquarters": ["ORGANIZATION:COUNTRY"],
        "org:stateorprovince_of_headquarters": ["ORGANIZATION:STATE_OR_PROVINCE"],
        "org:city_of_headquarters": [
            "ORGANIZATION:CITY",
            "ORGANIZATION:LOCATION",
        ],
        "org:shareholders": [
            "ORGANIZATION:PERSON",
            "ORGANIZATION:ORGANIZATION",
        ],
        "org:website": ["ORGANIZATION:URL"],
    }

    clf = NLIRelationClassifierWithMappingHead(
        labels=labels,
        template_mapping=template_mapping,
        pretrained_model="microsoft/deberta-v2-xlarge-mnli",
        valid_conditions=rules,
    )

    features = [
        REInputFeatures(
            subj="Billy Mays",
            obj="Tampa",
            pair_type="PERSON:CITY",
            context="Billy Mays, the bearded, boisterous pitchman who, as the undisputed king of TV yell and sell, became an unlikely pop culture icon, died at his home in Tampa, Fla, on Sunday",
            label="per:city_of_death",
        ),
        REInputFeatures(
            subj="Old Lane Partners",
            obj="Pandit",
            pair_type="ORGANIZATION:PERSON",
            context="Pandit worked at the brokerage Morgan Stanley for about 11 years until 2005, when he and some Morgan Stanley colleagues quit and later founded the hedge fund Old Lane Partners.",
            label="org:founded_by",
        ),
        # REInputFeatures(subj='He', obj='University of Maryland in College Park', context='He received an undergraduate degree from Morgan State University in 1950 and applied for admission to graduate school at the University of Maryland in College Park.', label='no_relation')
    ]

    predictions = clf(features, multiclass=True)
    for pred in predictions:
        pprint(sorted(list(zip(pred, labels)), reverse=True)[:5])



from . import NLITopicClassifierWithMappingHead

BABELDOMAINS_TOPICS = [
    "Animals",
    "Art, architecture, and archaeology",
    "Biology",
    "Business, economics, and finance",
    "Chemistry and mineralogy",
    "Computing",
    "Culture and society",
    "Education",
    "Engineering and technology",
    "Farming",
    "Food and drink",
    "Games and video games",
    "Geography and places",
    "Geology and geophysics",
    "Health and medicine",
    "Heraldry, honors, and vexillology",
    "History",
    "Language and linguistics",
    "Law and crime",
    "Literature and theatre",
    "Mathematics",
    "Media",
    "Meteorology",
    "Music",
    "Numismatics and currencies",
    "Philosophy and psychology",
    "Physics and astronomy",
    "Politics and government",
    "Religion, mysticism and mythology",
    "Royalty and nobility",
    "Sport and recreation",
    "Textile and clothing",
    "Transport and travel",
    "Warfare and defense",
]

BABELDOMAINS_TOPIC_MAPPING = {
    "Animals": "Animals",
    "Art": "Art, architecture, and archaeology",
    "Architecture": "Art, architecture, and archaeology",
    "Archaeology": "Art, architecture, and archaeology",
    "Art, architecture, and archaeology": "Art, architecture, and archaeology",
    "Biology": "Biology",
    "Business": "Business, economics, and finance",
    "Economics": "Business, economics, and finance",
    "Finance": "Business, economics, and finance",
    "Business, economics, and finance": "Business, economics, and finance",
    "Chemistry": "Chemistry and mineralogy",
    "Mineralogy": "Chemistry and mineralogy",
    "Chemistry and mineralogy": "Chemistry and mineralogy",
    "Computing": "Computing",
    "Culture": "Culture and society",
    "Society": "Culture and society",
    "Culture and society": "Culture and society",
    "Education": "Education",
    "Engineering": "Engineering and technology",
    "Technology": "Engineering and technology",
    "Engineering and technology": "Engineering and technology",
    "Farming": "Farming",
    "Food": "Food and drink",
    "Drink": "Food and drink",
    "Food and drink": "Food and drink",
    "Games": "Games and video games",
    "Video games": "Games and video games",
    "Games and video games": "Games and video games",
    "Geography": "Geography and places",
    "Places": "Geography and places",
    "Geography and places": "Geography and places",
    "Geology": "Geology and geophysics",
    "Geophysics": "Geology and geophysics",
    "Geology and geophysics": "Geology and geophysics",
    "Health": "Health and medicine",
    "Medicine": "Health and medicine",
    "Health and medicine": "Health and medicine",
    "Heraldry": "Heraldry, honors, and vexillology",
    "Honors": "Heraldry, honors, and vexillology",
    "Vexillology": "Heraldry, honors, and vexillology",
    "Heraldry, honors, and vexillology": "Heraldry, honors, and vexillology",
    "History": "History",
    "Language": "Language and linguistics",
    "Linguistics": "Language and linguistics",
    "Language and linguistics": "Language and linguistics",
    "Law": "Law and crime",
    "Crime": "Law and crime",
    "Law and crime": "Law and crime",
    "Literature": "Literature and theatre",
    "Theatre": "Literature and theatre",
    "Literature and theatre": "Literature and theatre",
    "Mathematics": "Mathematics",
    "Media": "Media",
    "Meteorology": "Meteorology",
    "Music": "Music",
    "Numismatics": "Numismatics and currencies",
    "Currencies": "Numismatics and currencies",
    "Numismatics and currencies": "Numismatics and currencies",
    "Philosophy": "Philosophy and psychology",
    "Psychology": "Philosophy and psychology",
    "Philosophy and psychology": "Philosophy and psychology",
    "Physics": "Physics and astronomy",
    "Astronomy": "Physics and astronomy",
    "Physics and astronomy": "Physics and astronomy",
    "Politics": "Politics and government",
    "Goverment": "Politics and government",
    "Politics and government": "Politics and government",
    "Religion": "Religion, mysticism and mythology",
    "Mysticism": "Religion, mysticism and mythology",
    "Mythology": "Religion, mysticism and mythology",
    "Religion, mysticism and mythology": "Religion, mysticism and mythology",
    "Royalty": "Royalty and nobility",
    "Nobility": "Royalty and nobility",
    "Royalty and nobility": "Royalty and nobility",
    "Sport": "Sport and recreation",
    "Recreation": "Sport and recreation",
    "Sport and recreation": "Sport and recreation",
    "Textile": "Textile and clothing",
    "Clothing": "Textile and clothing",
    "Textile and clothing": "Textile and clothing",
    "Transport": "Transport and travel",
    "Travel": "Transport and travel",
    "Transport and travel": "Transport and travel",
    "Warfare": "Warfare and defense",
    "Defense": "Warfare and defense",
    "Warfare and defense": "Warfare and defense",
}


class BabelDomainsClassifier(NLITopicClassifierWithMappingHead):
    """BabelDomainsClassifier

    Specific class for topic classification using BabelDomains topic set.
    """

    def __init__(self, **kwargs):
        super(BabelDomainsClassifier, self).__init__(
            pretrained_model="roberta-large-mnli",
            labels=BABELDOMAINS_TOPICS,
            topic_mapping=BABELDOMAINS_TOPIC_MAPPING,
            query_phrase="The domain of the sentence is about",
            entailment_position=2,
            **kwargs
        )



#!/usr/bin/env python
# coding=utf-8
# Copyright 2020 The HuggingFace Inc. team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Finetuning the library models for sequence classification on GLUE."""
# You can also adapt this script on your own text classification task. Pointers for this are left as comments.

import logging
import os
import random
import sys
from dataclasses import dataclass, field
from typing import Optional

import numpy as np
from datasets import load_dataset, load_metric

import transformers
from transformers import (
    AutoConfig,
    AutoModelForSequenceClassification,
    AutoTokenizer,
    DataCollatorWithPadding,
    EvalPrediction,
    HfArgumentParser,
    PretrainedConfig,
    Trainer,
    TrainingArguments,
    default_data_collator,
    set_seed,
)
from transformers.trainer_utils import get_last_checkpoint, is_main_process
from transformers.utils import check_min_version


# Will error if the minimal version of Transformers is not installed. Remove at your own risks.
check_min_version("4.5.0.dev0")

task_to_keys = {
    "cola": ("sentence", None),
    "mnli": ("premise", "hypothesis"),
    "mrpc": ("sentence1", "sentence2"),
    "qnli": ("question", "sentence"),
    "qqp": ("question1", "question2"),
    "rte": ("sentence1", "sentence2"),
    "sst2": ("sentence", None),
    "stsb": ("sentence1", "sentence2"),
    "wnli": ("sentence1", "sentence2"),
}

logger = logging.getLogger(__name__)


@dataclass
class DataTrainingArguments:
    """
    Arguments pertaining to what data we are going to input our model for training and eval.

    Using `HfArgumentParser` we can turn this class
    into argparse arguments to be able to specify them on
    the command line.
    """

    task_name: Optional[str] = field(
        default=None,
        metadata={"help": "The name of the task to train on: " + ", ".join(task_to_keys.keys())},
    )
    max_seq_length: int = field(
        default=128,
        metadata={
            "help": "The maximum total input sequence length after tokenization. Sequences longer "
            "than this will be truncated, sequences shorter will be padded."
        },
    )
    overwrite_cache: bool = field(
        default=False,
        metadata={"help": "Overwrite the cached preprocessed datasets or not."},
    )
    pad_to_max_length: bool = field(
        default=True,
        metadata={
            "help": "Whether to pad all samples to `max_seq_length`. "
            "If False, will pad the samples dynamically when batching to the maximum length in the batch."
        },
    )
    max_train_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": "For debugging purposes or quicker training, truncate the number of training examples to this "
            "value if set."
        },
    )
    max_val_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": "For debugging purposes or quicker training, truncate the number of validation examples to this "
            "value if set."
        },
    )
    max_test_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": "For debugging purposes or quicker training, truncate the number of test examples to this " "value if set."
        },
    )
    train_file: Optional[str] = field(
        default=None,
        metadata={"help": "A csv or a json file containing the training data."},
    )
    validation_file: Optional[str] = field(
        default=None,
        metadata={"help": "A csv or a json file containing the validation data."},
    )
    test_file: Optional[str] = field(
        default=None,
        metadata={"help": "A csv or a json file containing the test data."},
    )

    def __post_init__(self):
        if self.task_name is not None:
            self.task_name = self.task_name.lower()
            if self.task_name not in task_to_keys.keys():
                raise ValueError("Unknown task, you should pick one in " + ",".join(task_to_keys.keys()))
        elif self.train_file is None or self.validation_file is None:
            raise ValueError("Need either a GLUE task or a training/validation file.")
        else:
            train_extension = self.train_file.split(".")[-1]
            assert train_extension in [
                "csv",
                "json",
            ], "`train_file` should be a csv or a json file."
            validation_extension = self.validation_file.split(".")[-1]
            assert (
                validation_extension == train_extension
            ), "`validation_file` should have the same extension (csv or json) as `train_file`."


@dataclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.
    """

    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
    )
    config_name: Optional[str] = field(
        default=None,
        metadata={"help": "Pretrained config name or path if not the same as model_name"},
    )
    tokenizer_name: Optional[str] = field(
        default=None,
        metadata={"help": "Pretrained tokenizer name or path if not the same as model_name"},
    )
    cache_dir: Optional[str] = field(
        default=None,
        metadata={"help": "Where do you want to store the pretrained models downloaded from huggingface.co"},
    )
    use_fast_tokenizer: bool = field(
        default=True,
        metadata={"help": "Whether to use one of the fast tokenizer (backed by the tokenizers library) or not."},
    )
    model_revision: str = field(
        default="main",
        metadata={"help": "The specific model version to use (can be a branch name, tag name or commit id)."},
    )
    use_auth_token: bool = field(
        default=False,
        metadata={
            "help": "Will use the token generated when running `transformers-cli login` (necessary to use this script "
            "with private models)."
        },
    )


def main():
    # See all possible arguments in src/transformers/training_args.py
    # or by passing the --help flag to this script.
    # We now keep distinct sets of args, for a cleaner separation of concerns.

    parser = HfArgumentParser((ModelArguments, DataTrainingArguments, TrainingArguments))
    if len(sys.argv) == 2 and sys.argv[1].endswith(".json"):
        # If we pass only one argument to the script and it's the path to a json file,
        # let's parse it to get our arguments.
        model_args, data_args, training_args = parser.parse_json_file(json_file=os.path.abspath(sys.argv[1]))
    else:
        (
            model_args,
            data_args,
            training_args,
        ) = parser.parse_args_into_dataclasses()

    # Detecting last checkpoint.
    last_checkpoint = None
    if os.path.isdir(training_args.output_dir) and training_args.do_train and not training_args.overwrite_output_dir:
        last_checkpoint = get_last_checkpoint(training_args.output_dir)
        if last_checkpoint is None and len(os.listdir(training_args.output_dir)) > 0:
            raise ValueError(
                f"Output directory ({training_args.output_dir}) already exists and is not empty. "
                "Use --overwrite_output_dir to overcome."
            )
        elif last_checkpoint is not None:
            logger.info(
                f"Checkpoint detected, resuming training at {last_checkpoint}. To avoid this behavior, change "
                "the `--output_dir` or add `--overwrite_output_dir` to train from scratch."
            )

    # Setup logging
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
    )
    logger.setLevel(logging.INFO if is_main_process(training_args.local_rank) else logging.WARN)

    # Log on each process the small summary:
    logger.warning(
        f"Process rank: {training_args.local_rank}, device: {training_args.device}, n_gpu: {training_args.n_gpu}"
        + f"distributed training: {bool(training_args.local_rank != -1)}, 16-bits training: {training_args.fp16}"
    )
    # Set the verbosity to info of the Transformers logger (on main process only):
    if is_main_process(training_args.local_rank):
        transformers.utils.logging.set_verbosity_info()
        transformers.utils.logging.enable_default_handler()
        transformers.utils.logging.enable_explicit_format()
    logger.info(f"Training/evaluation parameters {training_args}")

    # Set seed before initializing model.
    set_seed(training_args.seed)

    # Get the datasets: you can either provide your own CSV/JSON training and evaluation files (see below)
    # or specify a GLUE benchmark task (the dataset will be downloaded automatically from the datasets Hub).
    #
    # For CSV/JSON files, this script will use as labels the column called 'label' and as pair of sentences the
    # sentences in columns called 'sentence1' and 'sentence2' if such column exists or the first two columns not named
    # label if at least two columns are provided.
    #
    # If the CSVs/JSONs contain only one non-label column, the script does single sentence classification on this
    # single column. You can easily tweak this behavior (see below)
    #
    # In distributed training, the load_dataset function guarantee that only one local process can concurrently
    # download the dataset.
    if data_args.task_name is not None:
        # Downloading and loading a dataset from the hub.
        datasets = load_dataset("glue", data_args.task_name)
    else:
        # Loading a dataset from your local files.
        # CSV/JSON training and evaluation files are needed.
        data_files = {
            "train": data_args.train_file,
            "validation": data_args.validation_file,
        }

        # Get the test dataset: you can provide your own CSV/JSON test file (see below)
        # when you use `do_predict` without specifying a GLUE benchmark task.
        if training_args.do_predict:
            if data_args.test_file is not None:
                train_extension = data_args.train_file.split(".")[-1]
                test_extension = data_args.test_file.split(".")[-1]
                assert (
                    test_extension == train_extension
                ), "`test_file` should have the same extension (csv or json) as `train_file`."
                data_files["test"] = data_args.test_file
            else:
                raise ValueError("Need either a GLUE task or a test file for `do_predict`.")

        for key in data_files.keys():
            logger.info(f"load a local file for {key}: {data_files[key]}")

        if data_args.train_file.endswith(".csv"):
            # Loading a dataset from local csv files
            datasets = load_dataset("csv", data_files=data_files)
        else:
            # Loading a dataset from local json files
            datasets = load_dataset("json", data_files=data_files)
    # See more about loading any type of standard or custom dataset at
    # https://huggingface.co/docs/datasets/loading_datasets.html.

    # Labels
    if data_args.task_name is not None:
        is_regression = data_args.task_name == "stsb"
        if not is_regression:
            label_list = datasets["train"].features["label"].names
            num_labels = len(label_list)
        else:
            num_labels = 1
    else:
        # Trying to have good defaults here, don't hesitate to tweak to your needs.
        is_regression = datasets["train"].features["label"].dtype in [
            "float32",
            "float64",
        ]
        if is_regression:
            num_labels = 1
        else:
            # A useful fast method:
            # https://huggingface.co/docs/datasets/package_reference/main_classes.html#datasets.Dataset.unique
            label_list = datasets["train"].unique("label")
            label_list.sort()  # Let's sort it for determinism
            num_labels = 3  # len(label_list) # WARNING: CHANGED

    # Load pretrained model and tokenizer
    #
    # In distributed training, the .from_pretrained methods guarantee that only one local process can concurrently
    # download model & vocab.
    config = AutoConfig.from_pretrained(
        model_args.config_name if model_args.config_name else model_args.model_name_or_path,
        num_labels=num_labels,
        finetuning_task=data_args.task_name,
        cache_dir=model_args.cache_dir,
        revision=model_args.model_revision,
        use_auth_token=True if model_args.use_auth_token else None,
    )
    tokenizer = AutoTokenizer.from_pretrained(
        model_args.tokenizer_name if model_args.tokenizer_name else model_args.model_name_or_path,
        cache_dir=model_args.cache_dir,
        use_fast=model_args.use_fast_tokenizer,
        revision=model_args.model_revision,
        use_auth_token=True if model_args.use_auth_token else None,
    )
    model = AutoModelForSequenceClassification.from_pretrained(
        model_args.model_name_or_path,
        from_tf=bool(".ckpt" in model_args.model_name_or_path),
        config=config,
        cache_dir=model_args.cache_dir,
        revision=model_args.model_revision,
        use_auth_token=True if model_args.use_auth_token else None,
    )

    # Preprocessing the datasets
    if data_args.task_name is not None:
        sentence1_key, sentence2_key = task_to_keys[data_args.task_name]
    else:
        # Again, we try to have some nice defaults but don't hesitate to tweak to your use case.
        non_label_column_names = [name for name in datasets["train"].column_names if name != "label"]
        if "sentence1" in non_label_column_names and "sentence2" in non_label_column_names:
            sentence1_key, sentence2_key = "sentence1", "sentence2"
        else:
            if len(non_label_column_names) >= 2:
                sentence1_key, sentence2_key = non_label_column_names[:2]
            else:
                sentence1_key, sentence2_key = non_label_column_names[0], None

    # Padding strategy
    if data_args.pad_to_max_length:
        padding = "max_length"
    else:
        # We will pad later, dynamically at batch creation, to the max sequence length in each batch
        padding = False

    # Some models have set the order of the labels to use, so let's make sure we do use it.
    label_to_id = None
    if (
        model.config.label2id != PretrainedConfig(num_labels=num_labels).label2id
        and data_args.task_name is not None
        and not is_regression
    ):
        # Some have all caps in their config, some don't.
        label_name_to_id = {k.lower(): v for k, v in model.config.label2id.items()}
        if list(sorted(label_name_to_id.keys())) == list(sorted(label_list)):
            label_to_id = {i: int(label_name_to_id[label_list[i]]) for i in range(num_labels)}
        else:
            logger.warn(
                "Your model seems to have been trained with labels, but they don't match the dataset: ",
                f"model labels: {list(sorted(label_name_to_id.keys()))}, dataset labels: {list(sorted(label_list))}."
                "\nIgnoring the model labels as a result.",
            )
    elif data_args.task_name is None and not is_regression:
        label_to_id = {v: i for i, v in enumerate(label_list)}

    if data_args.max_seq_length > tokenizer.model_max_length:
        logger.warn(
            f"The max_seq_length passed ({data_args.max_seq_length}) is larger than the maximum length for the"
            f"model ({tokenizer.model_max_length}). Using max_seq_length={tokenizer.model_max_length}."
        )
    max_seq_length = min(data_args.max_seq_length, tokenizer.model_max_length)

    def preprocess_function(examples):
        # Tokenize the texts
        args = (examples[sentence1_key],) if sentence2_key is None else (examples[sentence1_key], examples[sentence2_key])
        result = tokenizer(*args, padding=padding, max_length=max_seq_length, truncation=True)

        # Map labels to IDs (not necessary for GLUE tasks)
        if label_to_id is not None and "label" in examples:
            result["label"] = [(label_to_id[l] if l != -1 else -1) for l in examples["label"]]
        return result

    datasets = datasets.map(
        preprocess_function,
        batched=True,
        load_from_cache_file=not data_args.overwrite_cache,
    )
    if training_args.do_train:
        if "train" not in datasets:
            raise ValueError("--do_train requires a train dataset")
        train_dataset = datasets["train"]
        if data_args.max_train_samples is not None:
            train_dataset = train_dataset.select(range(data_args.max_train_samples))

    if training_args.do_eval:
        if "validation" not in datasets and "validation_matched" not in datasets:
            raise ValueError("--do_eval requires a validation dataset")
        eval_dataset = datasets["validation_matched" if data_args.task_name == "mnli" else "validation"]
        if data_args.max_val_samples is not None:
            eval_dataset = eval_dataset.select(range(data_args.max_val_samples))

    if training_args.do_predict or data_args.task_name is not None or data_args.test_file is not None:
        if "test" not in datasets and "test_matched" not in datasets:
            raise ValueError("--do_predict requires a test dataset")
        test_dataset = datasets["test_matched" if data_args.task_name == "mnli" else "test"]
        if data_args.max_test_samples is not None:
            test_dataset = test_dataset.select(range(data_args.max_test_samples))

    # Log a few random samples from the training set:
    if training_args.do_train:
        for index in random.sample(range(len(train_dataset)), 3):
            logger.info(f"Sample {index} of the training set: {train_dataset[index]}.")

    # Get the metric function
    if data_args.task_name is not None:
        metric = load_metric("glue", data_args.task_name)
    # TODO: When datasets metrics include regular accuracy, make an else here and remove special branch from
    # compute_metrics

    # You can define your custom compute_metrics function. It takes an `EvalPrediction` object (a namedtuple with a
    # predictions and label_ids field) and has to return a dictionary string to float.
    def compute_metrics(p: EvalPrediction):
        preds = p.predictions[0] if isinstance(p.predictions, tuple) else p.predictions
        preds = np.squeeze(preds) if is_regression else np.argmax(preds, axis=1)
        if data_args.task_name is not None:
            result = metric.compute(predictions=preds, references=p.label_ids)
            if len(result) > 1:
                result["combined_score"] = np.mean(list(result.values())).item()
            return result
        elif is_regression:
            return {"mse": ((preds - p.label_ids) ** 2).mean().item()}
        else:
            return {"accuracy": (preds == p.label_ids).astype(np.float32).mean().item()}

    # Data collator will default to DataCollatorWithPadding, so we change it if we already did the padding.
    if data_args.pad_to_max_length:
        data_collator = default_data_collator
    elif training_args.fp16:
        data_collator = DataCollatorWithPadding(tokenizer, pad_to_multiple_of=8)
    else:
        data_collator = None

    # Initialize our Trainer
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_dataset if training_args.do_train else None,
        eval_dataset=eval_dataset if training_args.do_eval else None,
        compute_metrics=compute_metrics,
        tokenizer=tokenizer,
        data_collator=data_collator,
    )

    # Training
    if training_args.do_train:
        checkpoint = None
        if last_checkpoint is not None:
            checkpoint = last_checkpoint
        elif os.path.isdir(model_args.model_name_or_path):
            # Check the config from that potential checkpoint has the right number of labels before using it as a
            # checkpoint.
            if AutoConfig.from_pretrained(model_args.model_name_or_path).num_labels == num_labels:
                checkpoint = model_args.model_name_or_path

        train_result = trainer.train(resume_from_checkpoint=checkpoint)
        metrics = train_result.metrics
        max_train_samples = data_args.max_train_samples if data_args.max_train_samples is not None else len(train_dataset)
        metrics["train_samples"] = min(max_train_samples, len(train_dataset))

        trainer.save_model()  # Saves the tokenizer too for easy upload

        trainer.log_metrics("train", metrics)
        trainer.save_metrics("train", metrics)
        trainer.save_state()

    # Evaluation
    if training_args.do_eval:
        logger.info("*** Evaluate ***")

        # Loop to handle MNLI double evaluation (matched, mis-matched)
        tasks = [data_args.task_name]
        eval_datasets = [eval_dataset]
        if data_args.task_name == "mnli":
            tasks.append("mnli-mm")
            eval_datasets.append(datasets["validation_mismatched"])

        for eval_dataset, task in zip(eval_datasets, tasks):
            metrics = trainer.evaluate(eval_dataset=eval_dataset)

            max_val_samples = data_args.max_val_samples if data_args.max_val_samples is not None else len(eval_dataset)
            metrics["eval_samples"] = min(max_val_samples, len(eval_dataset))

            trainer.log_metrics("eval", metrics)
            trainer.save_metrics("eval", metrics)

    if training_args.do_predict:
        logger.info("*** Test ***")

        # Loop to handle MNLI double evaluation (matched, mis-matched)
        tasks = [data_args.task_name]
        test_datasets = [test_dataset]
        if data_args.task_name == "mnli":
            tasks.append("mnli-mm")
            test_datasets.append(datasets["test_mismatched"])

        for test_dataset, task in zip(test_datasets, tasks):
            # Removing the `label` columns because it contains -1 and Trainer won't like that.
            test_dataset.remove_columns_("label")
            predictions = trainer.predict(test_dataset=test_dataset).predictions
            predictions = np.squeeze(predictions) if is_regression else np.argmax(predictions, axis=1)

            output_test_file = os.path.join(training_args.output_dir, f"test_results_{task}.txt")
            if trainer.is_world_process_zero():
                with open(output_test_file, "w") as writer:
                    logger.info(f"***** Test results {task} *****")
                    writer.write("index\tprediction\n")
                    for index, item in enumerate(predictions):
                        if is_regression:
                            writer.write(f"{index}\t{item:3.3f}\n")
                        else:
                            item = label_list[item]
                            writer.write(f"{index}\t{item}\n")


def _mp_fn(index):
    # For xla_spawn (TPUs)
    main()


if __name__ == "__main__":
    main()



from typing import Dict, List, Callable

from .base import ZeroaryTask, Features


class TopicClassificationFeatures(Features):
    """A class handler for the Topic Classification features. It inherits from `Features`."""

    pass


class IncorrectHypothesisTemplateError(Exception):
    """Raised when the `hypotesis_template` argument does not contain the '{label}' placeholder."""

    pass


class TopicClassificationTask(ZeroaryTask):
    """A class handler for Topic Classification task. It inherits from `ZeroaryTask` class."""

    def __init__(
        self,
        name: str = None,
        labels: List[str] = None,
        templates: Dict[str, List[str]] = None,
        hypothesis_template: str = "The domain of the sentence is about {label}.",
        features_class: type = TopicClassificationFeatures,
        preprocess_labels: bool = False,
        preprocess_fn: Callable = None,
        **kwargs,
    ) -> None:
        """Initialization of a TopicClassification task.

        Args:
            name (str, optional): A name for the task that may be used for to differentiate task when saving. Defaults to None.
            labels (List[str]): The labels for the task. Defaults to empty list.
            templates (Dict[str, List[str]], optional): The templates/verbalizations for the task. Defaults to None.
            hypothesis_template (str, optional): A meta template to generate hypothesis templates, if `templates` is None,
                then the templates will be the combinations of the `hypothesis_template` and the `labels`. It must contain the
                '{label}' placeholder. Defaults to "The domain of the sentence is about {label}.".
            features_class (type, optional): The `Features` class related to the task. Defaults to TopicClassificationFeatures.
            preprocess_labels (bool, optional): Whether to split the topic labels. Defaults to True.
            preprocess_fn (Callable, optional): The function that is applied if `split_labels` is True. If None then
                `TopicClassificationTask._split_labels_fn` is applied. Defaults to None.

        Raises:
            IncorrectHypothesisTemplateError: Raised when the `hypotesis_template` argument does not contain the '{label}' placeholder.
        """
        if not templates:

            if "{label}" not in hypothesis_template:
                raise IncorrectHypothesisTemplateError(
                    "The hypothesis_template argument must contain the '{label}' placeholder."
                )

            if preprocess_labels:
                split_labels_fn = preprocess_fn if preprocess_fn is not None else self._split_and_extend_labels_fn
                templates = {
                    label: [hypothesis_template.format(label=partial_label) for partial_label in split_labels_fn(label)]
                    for label in labels
                }
            else:
                templates = {label: [hypothesis_template.format(label=label)] for label in labels}

        super().__init__(
            name=name,
            required_variables=[],
            additional_variables=[],
            labels=labels,
            templates=templates,
            valid_conditions=None,
            features_class=features_class,
            **kwargs,
        )

    @staticmethod
    def _split_labels_fn(label: str) -> List[str]:
        labels = [
            partial_label.strip().capitalize()
            for partial_label in label.split(",")
            for partial_label in partial_label.split("and")
            if len(partial_label.strip())
        ]
        return list(set(labels))

    @staticmethod
    def _split_and_extend_labels_fn(label: str) -> List[str]:
        labels = [label] + [
            partial_label.strip().capitalize()
            for partial_label in label.split(",")
            for partial_label in partial_label.split("and")
            if len(partial_label.strip())
        ]
        return list(set(labels))



from typing import List

import numpy as np


class Dataset(list):
    """A simple class to handle the datasets.

    Inherits from `list`, so the instances should be added with `append` or `extend` methods to itself.
    """

    def __init__(self, labels: List[str], *args, **kwargs) -> None:
        """
        Args:
            labels (List[str]): The possible label set of the dataset.
        """
        super().__init__()

        self.labels2id = {label: i for i, label in enumerate(labels)}
        self.id2labels = {i: label for i, label in enumerate(labels)}

    @property
    def labels(self):
        # TODO: Unittest
        if not hasattr(self, "_labels"):
            self._labels = np.asarray([self.labels2id[inst.label] for inst in self])
        return self._labels



import os
import sys

import numpy as np
import torch


def np_softmax(x, dim=-1):
    e = np.exp(x)
    return e / np.sum(e, axis=dim, keepdims=True)


class TopicClassifier(object):
    def __init__(self, pretrained_model, topics, use_cuda=True, half=False, verbose=True):
        super().__init__()

        self.device = torch.device("cuda" if torch.cuda.is_available() and use_cuda else "cpu")
        self.topics = topics
        self.use_cuda = use_cuda
        self.half = half
        self.verbose = verbose

        # Supress stdout printing for model downloads
        if not verbose:
            sys.stdout = open(os.devnull, "w")
            self._initialize(pretrained_model)
            sys.stdout = sys.__stdout__
        else:
            self._initialize(pretrained_model)

        self.model.to(self.device)
        self.model.eval()
        if self.use_cuda and self.half and torch.cuda.is_available():
            self.model.half()

    def _initialize(self, pretrained_model):
        raise NotImplementedError

    def __call__(self, context, batch_size=1):
        raise NotImplementedError



import numpy as np
from sklearn.metrics import f1_score, precision_recall_fscore_support


def f1_score_(labels, preds, n_labels=42):
    return f1_score(labels, preds, labels=list(range(1, n_labels)), average="micro")


def precision_recall_fscore_(labels, preds, n_labels=42):
    p, r, f, _ = precision_recall_fscore_support(labels, preds, labels=list(range(1, n_labels)), average="micro")
    return p, r, f


def apply_threshold(output, threshold=0.0, ignore_negative_prediction=True):
    """Applies a threshold to determine whether is a relation or not"""
    output_ = output.copy()
    if ignore_negative_prediction:
        output_[:, 0] = 0.0
    activations = (output_ >= threshold).sum(-1).astype(np.int)
    output_[activations == 0, 0] = 1.00

    return output_.argmax(-1)


def find_optimal_threshold(labels, output, granularity=1000, metric=f1_score_):
    thresholds = np.linspace(0, 1, granularity)
    values = []
    for t in thresholds:
        preds = apply_threshold(output, threshold=t)
        values.append(metric(labels, preds))

    best_metric_id = np.argmax(values)
    best_threshold = thresholds[best_metric_id]

    return best_threshold, values[best_metric_id]



import sys
from pprint import pprint
from typing import List

import numpy as np
import torch
from tqdm import tqdm
from transformers import BertForNextSentencePrediction, AutoTokenizer

from a2t.base import Classifier


class NSPTopicClassifier(Classifier):
    def __init__(
        self,
        labels: List[str],
        *args,
        pretrained_model: str = "bert-large-uncased",
        use_cuda=True,
        query_phrase="Topic or domain about",
        positive_position=1,
        half=False,
        **kwargs,
    ):
        super().__init__(
            labels,
            pretrained_model,
            use_cuda=use_cuda,
            half=half,
            *args,
            **kwargs,
        )
        self.query_phrase = query_phrase
        self.cls_pos = positive_position

    def _initialize(self, pretrained_model):
        self.tokenizer = AutoTokenizer.from_pretrained(pretrained_model, use_fast=True)
        self.model = BertForNextSentencePrediction.from_pretrained(pretrained_model)

    def _run_batch(self, batch):
        with torch.no_grad():
            input_ids = self.tokenizer.batch_encode_plus(batch, padding=True)
            input_ids = torch.tensor(input_ids["input_ids"]).to(self.device)
            output = self.model(input_ids)[0][:, self.cls_pos].view(len(batch) // len(self.labels), -1)
            output = torch.softmax(output, dim=-1).detach().cpu().numpy()

        return output

    def __call__(self, contexts, batch_size=1):
        if not isinstance(contexts, list):
            contexts = [contexts]

        batch, outputs = [], []
        for i, context in tqdm(enumerate(contexts), total=len(contexts)):
            sentences = [f'{context} {self.tokenizer.sep_token} {self.query_phrase} "{topic}".' for topic in self.labels]
            batch.extend(sentences)

            if (i + 1) % batch_size == 0:
                output = self._run_batch(batch)
                outputs.append(output)
                batch = []

        if len(batch) > 0:
            output = self._run_batch(batch)
            outputs.append(output)

        outputs = np.vstack(outputs)

        return outputs


if __name__ == "__main__":

    if len(sys.argv) < 2:
        print("Usage:\tpython3 get_topics.py topics.txt input_file.txt\n\tpython3 get_topics.py topics.txt < " "input_file.txt")
        exit(1)

    with open(sys.argv[1], "rt") as f:
        topics = [topic.rstrip().replace("_", " ") for topic in f]

    input_stream = open(sys.argv[2], "rt") if len(sys.argv) == 3 else sys.stdin

    clf = NSPTopicClassifier("bert-large-uncased", labels=topics)

    for line in input_stream:
        line = line.rstrip()
        output_probs = clf(line)[0]
        topic_dist = sorted(list(zip(output_probs, topics)), reverse=True)
        print(line)
        pprint(topic_dist)
        print()



import unittest

from a2t.data.tacred import TACREDRelationClassificationDataset
from a2t.tasks.tuple_classification import TACREDRelationClassificationTask, TACREDFeatures

import os


class TestTACREDRelationClassificationDataset(unittest.TestCase):
    @unittest.skipIf(not os.path.exists("data/tacred/dev.json"), "No TACRED data available at data/tacred/")
    def test_data_loader(self):

        TACRED_LABELS = [
            "no_relation",
            "org:alternate_names",
            "org:city_of_headquarters",
            "org:country_of_headquarters",
            "org:dissolved",
            "org:founded",
            "org:founded_by",
            "org:member_of",
            "org:members",
            "org:number_of_employees/members",
            "org:parents",
            "org:political/religious_affiliation",
            "org:shareholders",
            "org:stateorprovince_of_headquarters",
            "org:subsidiaries",
            "org:top_members/employees",
            "org:website",
            "per:age",
            "per:alternate_names",
            "per:cause_of_death",
            "per:charges",
            "per:children",
            "per:cities_of_residence",
            "per:city_of_birth",
            "per:city_of_death",
            "per:countries_of_residence",
            "per:country_of_birth",
            "per:country_of_death",
            "per:date_of_birth",
            "per:date_of_death",
            "per:employee_of",
            "per:origin",
            "per:other_family",
            "per:parents",
            "per:religion",
            "per:schools_attended",
            "per:siblings",
            "per:spouse",
            "per:stateorprovince_of_birth",
            "per:stateorprovince_of_death",
            "per:stateorprovinces_of_residence",
            "per:title",
        ]

        # Test class creation without errors
        dataset = TACREDRelationClassificationDataset("data/tacred/dev.json", labels=TACRED_LABELS)

        # Test instance class is the correct
        self.assertTrue(isinstance(dataset[0], TACREDFeatures))

        # Test that instances are loaded correctly
        self.assertTrue(dataset[0].label == "per:title")

        # Test that above is true for all the instances
        task = TACREDRelationClassificationTask(TACRED_LABELS, {"per:title": ["{subj} is also known as {obj}"]}, None)

        task.assert_features_class(dataset)



from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer,
    AutoConfig,
)
from transformers.optimization import get_linear_schedule_with_warmup
import argparse
import json
import gzip
import torch
from torch.utils.data import TensorDataset, DataLoader
from torch.optim import SGD, AdamW, Adam

# from apex import amp
from tqdm import tqdm
import numpy as np
from pprint import pprint
import os

try:
    from apex import amp
except:
    pass


class EarlyStopping(object):
    """Simple implementation of Early Stopping concept based on the validation loss.

    Usage:
    ```
    >>> early_stopping = EarlyStopping(patience=3, delta=.001)

    >>> for epoch in epochs:
    >>>     val_loss = ..
    >>>     if early_stopping(val_loss):
    >>>         break
    ```
    """

    def __init__(self, patience=3, delta=0.0, save_checkpoint_fn=None):
        super().__init__()

        self.patience = patience
        self.delta = delta
        self.save_checkpoint_fn = save_checkpoint_fn

        self.best_loss = np.inf
        self.best_epoch = -1
        self.steps = 0
        self.epoch = -1

    def __call__(self, val_loss: float, *args, **kwargs) -> bool:
        self.epoch += 1
        diff = (self.best_loss + self.delta) - val_loss
        if diff > 0:
            self.steps = self.patience
            if diff > self.delta:
                self.best_epoch = self.epoch
                self.best_loss = val_loss
                if self.save_checkpoint_fn:
                    self.save_checkpoint_fn(*args, val_loss=self.best_loss, **kwargs)
        else:
            self.steps -= 1

        return self.steps <= 0

    def get_best_epoch(self):
        return self.best_epoch


def parse_args():

    parser = argparse.ArgumentParser(
        "finetune_classifier",
        usage="finetune_classifier train_data eval_data --config config.json",
    )
    parser.add_argument("--config", type=str, dest="config", help="Training configuration")

    args = parser.parse_args()

    return args


def train(opt):

    # Load the configuration
    with open(opt.config, "rt") as f:
        config = json.load(f)

    os.makedirs(config["output_path"], exist_ok=True)

    # Load the topics/labels
    with open(config["topics"], "rt") as f:
        topics = [topic.rstrip().replace("_", " ") for topic in f]
        topic2id = {topic: i for i, topic in enumerate(topics)}

    # Load the sinset-->gloss mapping
    glosses = {}
    with gzip.open(config["glosses"], "rt") as f:
        for line in f:
            row = line.split()
            idx, gloss = row[0], " ".join(row[1:]).replace("\t", " ")
            glosses[idx] = gloss

    # Load train and eval data
    train_data, train_labels, eval_data, eval_labels = [], [], [], []
    with open(config["eval_data"], "rt") as f:
        for line in f:
            idx, label, _ = line.strip().split("\t")
            eval_data.append(idx)
            eval_labels.append(topic2id[label])

    with gzip.open(config["train_data"], "rt") as f:
        for line in f:
            row = line.strip().split("\t")
            if row[0] in glosses and float(row[-1]) > config["minimum_confidence"] and row[0] not in eval_data:
                train_data.append(row[0])
                train_labels.append(topic2id[row[1]])

    # Convert train and eval data to glosses
    train_glosses = [glosses[instance] for instance in train_data]
    eval_glosses = [glosses[instance] for instance in eval_data]

    # Define the model and tokenizer
    tokenizer = AutoTokenizer.from_pretrained(config["pretrained_model"])
    tokenizer.save_pretrained(config["output_path"])
    cfg = AutoConfig.from_pretrained(config["pretrained_model"])
    cfg.num_labels = len(topics)
    cfg.label2id = topic2id
    cfg.id2label = {str(idx): label for label, idx in topic2id.items()}
    model = AutoModelForSequenceClassification.from_pretrained(config["pretrained_model"], config=cfg)

    # Prepare data for training
    train_dataset = tokenizer(
        train_glosses,
        padding=True,
        truncation=True,
        max_length=config["max_length"],
    )
    eval_dataset = tokenizer(
        eval_glosses,
        padding=True,
        truncation=True,
        max_length=config["max_length"],
    )

    train_dataset = TensorDataset(
        torch.tensor(train_dataset["input_ids"]),
        torch.tensor(train_dataset["attention_mask"]),
        torch.tensor(train_labels),
    )
    eval_dataset = TensorDataset(
        torch.tensor(eval_dataset["input_ids"]),
        torch.tensor(eval_dataset["attention_mask"]),
        torch.tensor(eval_labels),
    )

    train_dataloader = DataLoader(train_dataset, batch_size=config["batch_size"], shuffle=True)
    eval_dataloader = DataLoader(eval_dataset, batch_size=config["eval_batch_size"])

    def get_parameters(model, freeze=False):
        param_optimizer = list(model.named_parameters())
        if freeze:
            for n, p in param_optimizer:
                if "classifier" not in n:
                    p.requires_grad = False
        no_decay = ["bias", "LayerNorm.bias", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)],
                "weight_decay": 0.01,
            },
            {
                "params": [p for n, p in param_optimizer if any(nd in n for nd in no_decay)],
                "weight_decay": 0.0,
            },
        ]

        return optimizer_grouped_parameters

    optimizer = AdamW(get_parameters(model, config["freeze"]), lr=config["learning_rate"])

    model.cuda()
    if config["half"]:
        model, optimizer = amp.initialize(model, optimizer, opt_level="O2", keep_batchnorm_fp32=True)

    def save_checkpoint_fn(model, output_path, **kwargs):
        model.save_pretrained(output_path)

    config["training_log"] = {}

    early_stopping = EarlyStopping(patience=config["patience"], save_checkpoint_fn=save_checkpoint_fn)
    for epoch in range(config["epochs"]):
        config["training_log"][f"epoch_{epoch}"] = {}
        model.train()
        total_loss, accuracy = 0.0, 0.0
        correct, total = 0, 0
        progress = tqdm(
            enumerate(train_dataloader),
            desc=f"Epoch: {epoch} - Loss: {total_loss} - Accuracy: {accuracy}",
            total=len(train_dataset) // config["batch_size"],
        )
        for i, batch in progress:
            # Batch to cuda
            input_ids, attention_mask, labels = batch

            loss, logits = model(
                input_ids=input_ids.cuda(),
                attention_mask=attention_mask.cuda(),
                labels=labels.cuda(),
            )

            if config["half"]:
                with amp.scale_loss(loss, optimizer) as scale_loss:
                    scale_loss.backward()
            else:
                loss.backward()

            optimizer.step()
            optimizer.zero_grad()

            total_loss += loss.item()
            correct += sum(np.argmax(logits.detach().cpu().numpy(), -1) == labels.numpy())
            total += input_ids.shape[0]
            accuracy = correct / total

            progress.set_description(f"Epoch: {epoch} - Loss: {total_loss/(i+1):.3f} - Accuracy: {accuracy:.3f}")

        config["training_log"][f"epoch_{epoch}"]["train_loss"] = total_loss / (i + 1)
        config["training_log"][f"epoch_{epoch}"]["train_accuracy"] = correct / total

        model.eval()
        with torch.no_grad():
            total_loss, accuracy = 0.0, 0.0
            correct, total = 0, 0
            progress = tqdm(
                enumerate(eval_dataloader),
                desc=f"Epoch: {epoch} - Loss: {total_loss} - Accuracy: {accuracy}",
                total=len(eval_dataset) // config["eval_batch_size"],
            )
            for i, batch in progress:
                # Batch to cuda
                input_ids, attention_mask, labels = batch

                loss, logits = model(
                    input_ids=input_ids.cuda(),
                    attention_mask=attention_mask.cuda(),
                    labels=labels.cuda(),
                )

                total_loss += loss.item()
                correct += sum(np.argmax(logits.detach().cpu().numpy(), -1) == labels.numpy())
                total += input_ids.shape[0]
                accuracy = correct / total

                progress.set_description(f"Epoch: {epoch} - Loss: {total_loss/(i+1):.3f} - Accuracy: {accuracy:.3f}")

        config["training_log"][f"epoch_{epoch}"]["eval_loss"] = total_loss / (i + 1)
        config["training_log"][f"epoch_{epoch}"]["eval_accuracy"] = correct / total

        if early_stopping(total_loss / (i + 1), model, output_path=config["output_path"]):
            break

    # Save the new config file
    with open(opt.config, "wt") as f:
        json.dump(config, f, indent=4)


if __name__ == "__main__":
    opt = parse_args()
    train(opt)



from collections import defaultdict
import inspect
import warnings
from dataclasses import dataclass, field, fields
from typing import Callable, List, Dict, Union
import re
import json
import os

import numpy as np
from sklearn.metrics import accuracy_score, precision_recall_fscore_support

from a2t.utils import find_optimal_threshold, apply_threshold


@dataclass
class Features:
    """A simple class to handle the features information.

    Args:
        context (str): The context sentence.
        label (str, optional): The label of the instance.
        inst_type (str, optional): The type of the instance. This information is used for the `valid_conditions' constraints.
    """

    context: str
    label: str = None
    inst_type: str = None


class IncorrectFeatureTypeError(Exception):
    pass


@dataclass
class Task:
    """Abstract class for Tasks definition.

    The method `_assert_constraints()` must be overrided.

    Args:
        name (str, optional): A name for the task that may be used for to differentiate task when saving. Defaults to None.
        required_variables (List[str], optional): The variables required to perform the task and must be implemented by the `Features` class. Defaults to empty list.
        additional_variables (List[str], optional): The variables not required to perform the task and must be implemented by the `Features` class. Defaults to empty list.
        labels (List[str], optional): The labels for the task. Defaults to empty list.
        templates (Dict[str, List[str]], optional): The templates/verbalizations for the task. Defaults to empty dict.
        valid_conditions (Dict[str, List[str]], optional): The valid conditions or constraints for the task.
        negative_label_id (int, optional): The index of the negative label or -1 if no negative label exist. A negative label is for example the class `Other` on NER, that means that the specific token is not a named entity. Defaults to -1.
        multi_label (bool, optional): Wheter the task must be treated as multi-label or not. You should treat as multi-label task a task that contains a negative label. Defaults to False.
        features_class (type, optional): The `Features` class related to the task. Default to `Features`.
    """

    name: str = None
    required_variables: List[str] = field(default_factory=list)
    additional_variables: List[str] = field(default_factory=list)
    labels: List[str] = field(default_factory=list)
    templates: Dict[str, List[str]] = field(default_factory=dict)
    valid_conditions: Dict[str, List[str]] = field(default_factory=dict)
    negative_label_id: int = -1  # -1 for no negative class
    multi_label: bool = False
    features_class: type = Features

    def __post_init__(self):
        self._assert_minimal_constraints()
        self._assert_constraints()

        self.label2id = {label: i for i, label in enumerate(self.labels)}
        self.n_labels = len(self.labels)

        if not self.templates:
            self.templates = {}

        # Create the templates to label mapping
        self.template2label = defaultdict(list)
        for label, templates in self.templates.items():
            for template in templates:
                self.template2label[template].append(label)

        self.template_list = list(self.template2label.keys())
        template2id = {template: i for i, template in enumerate(self.template_list)}

        self.label2templateid = defaultdict(list)
        for label, templates in self.templates.items():
            self.label2templateid[label].extend([template2id[template] for template in templates])

        # Create the valid_conditions matrix
        if self.valid_conditions:
            self._valid_conditions = {}
            self._always_valid_labels = np.zeros(self.n_labels)
            self._always_valid_labels[self.negative_label_id] = 1.0
            for label, conditions in self.valid_conditions.items():
                if label not in self.labels:
                    continue
                for condition in conditions:
                    if condition == "*":
                        self._always_valid_labels[self.label2id[label]] = 1.0
                        continue
                    if condition not in self._valid_conditions:
                        self._valid_conditions[condition] = np.zeros(self.n_labels)
                        if self.negative_label_id >= 0:
                            self._valid_conditions[condition][self.negative_label_id] = 1.0
                    self._valid_conditions[condition][self.label2id[label]] = 1.0
        else:
            self._valid_conditions = None

        def idx2label(idx):
            return self.labels[idx]

        self.idx2label = np.vectorize(idx2label)

    def __repr__(self) -> str:
        class_name = self.name if self.name else str(self.__class__)
        labels_repr = self.labels.__repr__()
        if len(labels_repr) > 89:
            labels_repr = self.labels[:3].__repr__().replace("]", ", ...]")
        templates_repr = len(self.template2label)
        feature_class_repr = str(self.features_class)

        return (
            f"{class_name} ("
            f"\n\tLabels: {labels_repr}"
            f"\n\tTemplates: {templates_repr}"
            f"\n\tFeatures: {feature_class_repr}"
            "\n)"
        )

    def _assert_constraints(self):
        raise NotImplementedError(f"{self.__class__} is an abstract class. This method should be implemented.")

    def _assert_minimal_constraints(self):
        assert len(self.labels) > 0, "The number of labels should be greather than 0."

        assert self.negative_label_id < len(
            self.labels
        ), "The id for the negative label should be lower than the amount of labels."

        if self.negative_label_id >= 0:
            assert self.templates is not None and len(
                [value for values in self.templates.values() for value in values]
            ), "`templates` parameter must not be None nor empty."

        assert all(
            key in self.labels for key in self.templates.keys()
        ), "All the keys of templates dicts must be defined on labels."

        if self.valid_conditions:
            assert all(
                key in self.labels for key in self.valid_conditions.keys()
            ), "All the keys of valid_conditions dict must be defined on labels."

        assert all(
            var in self.features_class.__dict__["__dataclass_fields__"]
            for var in self.required_variables + self.additional_variables
        ), "All variables should be defined on the features_class."

        assert all(
            var.strip("{").strip("}") in [*self.required_variables, *self.additional_variables]
            for templates in self.templates.values()
            for template in templates
            for var in re.findall(r"{\w+}", template)
        )

    def assert_features_class(self, features: List[Features]) -> None:
        """Assert that all features are instance of the task specific `Features` class.

        Args:
            features (List[Features]): The list of features to check.

        Raises:
            IncorrectFeatureTypeError: Raised when any feature is not an instance of the task specific `Features` class.
        """
        for feature in features:
            if not isinstance(feature, self.features_class):
                raise IncorrectFeatureTypeError(
                    f"Incorrect feature type given. Expected {self.features_class} but obtained {type(feature)}."
                )

    def generate_premise_hypotheses_pairs(self, features: List[Features], sep_token: str = "</s>") -> List[str]:
        """Generate premise-hypothesis pairs based on the `Task` templates.

        Args:
            features (List[Features]): The list of features.
            sep_token (str, optional): The model specific separator token. Defaults to "</s>".

        Returns:
            List[str]: The list of premise-hypothesis pairs generated from the features and templates.
        """
        if not isinstance(features, list):
            features = [features]

        sentence_pairs = [
            f"{feature.context} {sep_token} {template.format(**feature.__dict__)}"
            for feature in features
            for template in self.template_list
        ]
        return sentence_pairs

    def reverse_to_labels(self, template_probs: np.ndarray, collate_fn: Callable = np.max) -> np.ndarray:
        """A function that maps template probabilities to label probabilites. By default, the maximum probabilities among
        label related templates is used.

        Args:
            template_probs (np.ndarray): (batch_size, n_templates) The templates probabilites.
            collate_fn (Callable, optional): The probabilites collate function. Defaults to np.max.

        Returns:
            np.ndarray: (batch_size, n_labels) The labels probabilities.
        """
        outputs = np.hstack(
            [
                collate_fn(template_probs[:, self.label2templateid[label]], axis=-1, keepdims=True)
                if label in self.label2templateid
                else np.zeros((template_probs.shape[0], 1))
                for label in self.labels
            ]
        )
        return outputs

    def apply_valid_conditions(self, features: List[Features], probs: np.ndarray) -> np.ndarray:
        """Applies the valid conditions to the labels probabilities. If a constraint is not satisfied the probability is set to 0.

        Args:
            features (List[Features]): (batch_size,) The list of features.
            probs (np.ndarray): (batch_size, n_labels) The labels probabilities.

        Returns:
            np.ndarray: (batch_size, n_labels) The labels probabilities.
        """
        if self._valid_conditions:
            mask_matrix = np.stack(
                [self._valid_conditions.get(feature.inst_type, np.zeros(self.n_labels)) for feature in features],
                axis=0,
            )
            probs = probs * np.logical_or(mask_matrix, self._always_valid_labels)  # TODO: Need a test
        return probs

    def compute_metrics(
        self, labels: np.ndarray, output: np.ndarray, threshold: Union[str, float] = "optimize"
    ) -> Dict[str, float]:
        """Compute the metrics for the given task. This method is abstract and needs to be overrided.

        Args:
            labels (np.ndarray): (batch_size,) The correct labels.
            output (np.ndarray): (batch_size, n_labels) The labels probabilities.
            threshold (Union[str, float], optional): The threshold to use on the evaluation. Options:

                * **"default"**: The threshold is set to 0.5.
                * **"optimize"**: Optimize the threshold with the `labels`. Intended to be used on the development split.
                * **`float`**: A specific float value for the threshold.

                Defaults to "optimize".

        Raises:
            NotImplementedError: Raise if not overrided.

        Returns:
            Dict[str, float]: Dict with the resulting metrics.
        """
        # TODO: Unittest
        raise NotImplementedError("This method must be implemented.")

    @classmethod
    def from_config(cls, file_path: str) -> object:
        """Loads the Task instance from a configuration file.

        Args:
            file_path (str): The path to the configuration file.

        Returns:
            Task: A `Task` instance based on the configuration file.
        """
        with open(file_path, "rt") as f:
            config = json.load(f)

        if "features_class" in config:
            components = config["features_class"].split(".")
            mod = __import__(components[0])
            for comp in components[1:]:
                mod = getattr(mod, comp)

            config["features_class"] = mod

        params = set([p.name for p in fields(cls)]) | set(inspect.signature(cls).parameters.keys())
        params = {key: config[key] for key in params if key in config}

        return cls(**params)

    def to_config(self, file_path: str) -> None:
        """Saves the task instance to a configuration file.

        Args:
            file_path (str): The path to the configuration file.
        """
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "wt") as f:
            values = {key: value for key, value in vars(self).items()}
            values["features_class"] = values["features_class"].__module__ + "." + values["features_class"].__name__
            for key in ["label2id", "idx2label", "n_labels", "template2label", "label2templateid", "_valid_conditions"]:
                del values[key]

            json.dump(values, f, indent=4)


class ZeroaryFeatures(Features):
    """A features class for `ZeroaryTask`. It only requires a `context` argument."""

    pass


@dataclass
class ZeroaryTask(Task):
    """A `Task` implementation for Text Classification like tasks.

    Args:
        name (str, optional): A name for the task that may be used for to differentiate task when saving. Defaults to None.
        additional_variables (List[str], optional): The variables not required to perform the task and must be implemented by the `Features` class. Defaults to empty list.
        labels (List[str], optional): The labels for the task. Defaults to empty list.
        templates (Dict[str, List[str]], optional): The templates/verbalizations for the task. Defaults to empty dict.
        valid_conditions (Dict[str, List[str]], optional): The valid conditions or constraints for the task.
        multi_label (bool, optional): Wheter the task must be treated as multi-label or not. You should treat as multi-label task a task that contains a negative label. Defaults to False.
        features_class (type, optional): The `Features` class related to the task. Default to `ZeroaryFeatures`.
    """

    features_class: type = ZeroaryFeatures

    def _assert_constraints(self):
        # Assert the number of required variables to be 0
        assert len(self.required_variables) == 0, "Zero-ary tasks like Text classifiation do not require any variable."

    def compute_metrics(self, labels: np.ndarray, output: np.ndarray, threshold: Union[str, float] = None) -> Dict[str, float]:
        """Compute the metrics for the given task. By default on `ZeroaryTask` the Accuracy is computed.

        Args:
            labels (np.ndarray): (batch_size,) The correct labels.
            output (np.ndarray): (batch_size, n_labels) The labels probabilities.
            threshold (Union[str, float], optional): No threshold is needed on `ZeroaryTask`.

        Returns:
            Dict[str, float]: Dict with the resulting metrics.
        """
        # TODO: Unittest
        if threshold:
            warnings.warn(f"{self.__class__} do not require 'threshold', ignored.")

        return {"accuracy_score": accuracy_score(labels, output.argmax(-1))}


@dataclass
class UnaryFeatures(Features):
    """A features class for `UnaryTask`. It requires `context` and `X` arguments."""

    X: str = None


@dataclass
class UnaryTask(Task):
    """A `Task` implementation for Span Classification like tasks.

    Args:
        name (str, optional): A name for the task that may be used for to differentiate task when saving. Defaults to None.
        additional_variables (List[str], optional): The variables not required to perform the task and must be implemented by the `Features` class. Defaults to empty list.
        labels (List[str], optional): The labels for the task. Defaults to empty list.
        templates (Dict[str, List[str]], optional): The templates/verbalizations for the task. Defaults to empty dict.
        valid_conditions (Dict[str, List[str]], optional): The valid conditions or constraints for the task.
        multi_label (bool, optional): Wheter the task must be treated as multi-label or not. You should treat as multi-label task a task that contains a negative label. Defaults to False.
        features_class (type, optional): The `Features` class related to the task. Default to `UnaryFeatures`.
    """

    features_class: type = UnaryFeatures

    def _assert_constraints(self):
        # Assert the number of required variables to be 1
        assert len(self.required_variables) == 1, "Unary-ary tasks like Span classifiation requires 1 variable."

    def compute_metrics(self, labels: np.ndarray, output: np.ndarray, threshold: Union[str, float] = "optimize"):
        """Compute the metrics for the given task. By default on `UnaryTask` the Accuracy is computed if
        the `negative_label_id` is < 0, otherwise the Precision, Recall, F1-Score and positive Accuracy are
        computed.

        Args:
            labels (np.ndarray): (batch_size,) The correct labels.
            output (np.ndarray): (batch_size, n_labels) The labels probabilities.
            threshold (Union[str, float], optional): The threshold to use on the evaluation. Options:

                * "default": The threshold is set to 0.5.
                * "optimize": Optimize the threshold with the `labels`. Intended to be used on the development split.
                * `float`: A specific float value for the threshold.

                Defaults to "optimize".

        Returns:
            Dict[str, float]: Dict with the resulting metrics.
        """
        # TODO: Unittest
        if threshold not in ["default", "optimize"] and not isinstance(threshold, float):
            raise ValueError("Threshold must be either 'default', 'optimize' or a float value.")

        if threshold == "default":
            threshold = 0.5

        if threshold == "optimize":
            threshold, _ = find_optimal_threshold(labels, output, negative_label_id=self.negative_label_id)

        results = {"optimal_threshold": threshold}
        if self.negative_label_id < 0:
            results["accuracy_score"] = accuracy_score(labels, output.argmax(-1))
        else:
            output_ = apply_threshold(output, threshold=threshold, negative_label_id=self.negative_label_id)
            positive_labels = list(set(range(len(self.labels))) - set([self.negative_label_id]))
            output_pos = output.copy()
            output_pos[:, self.negative_label_id] = 0.0

            results["positive_accuracy"] = accuracy_score(
                labels[labels != self.negative_label_id], output_pos[labels != self.negative_label_id, :].argmax(-1)
            )

            pre, rec, f1, _ = precision_recall_fscore_support(labels, output_, labels=positive_labels, average="micro")
            results["precision"] = pre
            results["recall"] = rec
            results["f1-score"] = f1

        return results


@dataclass
class BinaryFeatures(Features):
    """A features class for `BinaryTask`. It requires `context`, `X` and `Y` arguments."""

    X: str = None
    Y: str = None


@dataclass
class BinaryTask(Task):
    """A `Task` implementation for Relation Classification like tasks.

    Args:
        name (str, optional): A name for the task that may be used for to differentiate task when saving. Defaults to None.
        additional_variables (List[str], optional): The variables not required to perform the task and must be implemented by the `Features` class. Defaults to empty list.
        labels (List[str], optional): The labels for the task. Defaults to empty list.
        templates (Dict[str, List[str]], optional): The templates/verbalizations for the task. Defaults to empty dict.
        valid_conditions (Dict[str, List[str]], optional): The valid conditions or constraints for the task.
        multi_label (bool, optional): Wheter the task must be treated as multi-label or not. You should treat as multi-label task a task that contains a negative label. Defaults to False.
        features_class (type, optional): The `Features` class related to the task. Default to `BinaryFeatures`.
    """

    features_class: type = BinaryFeatures

    def _assert_constraints(self):
        # Assert the number of required variables to be 2
        assert len(self.required_variables) == 2, "Binary-ary tasks like Tuple classifiation require 2 variable."

    def compute_metrics(self, labels: np.ndarray, output: np.ndarray, threshold: Union[str, float] = "optimize"):
        """Compute the metrics for the given task. By default on `BinaryTask` the Accuracy is computed if
        the `negative_label_id` is < 0, otherwise the Precision, Recall, F1-Score and positive Accuracy are
        computed.

        Args:
            labels (np.ndarray): (batch_size,) The correct labels.
            output (np.ndarray): (batch_size, n_labels) The labels probabilities.
            threshold (Union[str, float], optional): The threshold to use on the evaluation. Options:

                * "default": The threshold is set to 0.5.
                * "optimize": Optimize the threshold with the `labels`. Intended to be used on the development split.
                * `float`: A specific float value for the threshold.

                Defaults to "optimize".

        Returns:
            Dict[str, float]: Dict with the resulting metrics.
        """

        # TODO: Unittest + documentation
        if threshold not in ["default", "optimize"] and not isinstance(threshold, float):
            raise ValueError("Threshold must be either 'default', 'optimize' or a float value.")

        if threshold == "default":
            threshold = 0.5

        if threshold == "optimize":
            threshold, _ = find_optimal_threshold(labels, output, negative_label_id=self.negative_label_id)

        results = {"optimal_threshold": threshold}
        if self.negative_label_id < 0:
            results["accuracy_score"] = accuracy_score(labels, output.argmax(-1))
        else:
            output_ = apply_threshold(output, threshold=threshold, negative_label_id=self.negative_label_id)
            positive_labels = list(set(range(len(self.labels))) - set([self.negative_label_id]))
            output_pos = output.copy()
            output_pos[:, self.negative_label_id] = 0.0

            results["positive_accuracy"] = accuracy_score(
                labels[labels != self.negative_label_id], output_pos[labels != self.negative_label_id, :].argmax(-1)
            )

            pre, rec, f1, _ = precision_recall_fscore_support(labels, output_, labels=positive_labels, average="micro")
            results["precision"] = pre
            results["recall"] = rec
            results["f1-score"] = f1

        return results



from typing import Dict, List
from dataclasses import dataclass

from .text_classification import IncorrectHypothesisTemplateError
from .base import UnaryTask, Features


@dataclass
class NamedEntityClassificationFeatures(Features):
    X: str = None


class NamedEntityClassificationTask(UnaryTask):
    """A class handler for Named Entity Classification task. It inherits from `UnaryTask` class.

    TODO: Add documentation.

    """

    def __init__(
        self,
        name: str,
        labels: List[str],
        *args,
        required_variables: List[str] = ["X"],
        additional_variables: List[str] = ["inst_type"],
        templates: Dict[str, List[str]] = None,
        valid_conditions: Dict[str, List[str]] = None,
        hypothesis_template: str = "{X} is a {label}.",
        features_class: type = NamedEntityClassificationFeatures,
        multi_label: bool = True,
        negative_label_id: int = 0,
        **kwargs
    ) -> None:
        """Initialization of a NamedEntityClassificationTask task.

        Args:
            name (str): [description]
            labels (List[str]): [description]
            required_variables (List[str], optional): [description]. Defaults to ["X"].
            additional_variables (List[str], optional): [description]. Defaults to ["inst_type"].
            templates (Dict[str, List[str]], optional): [description]. Defaults to None.
            valid_conditions (Dict[str, List[str]], optional): [description]. Defaults to None.
            hypothesis_template (str, optional): [description]. Defaults to "{X} is a {label}.".
            features_class (type, optional): [description]. Defaults to NamedEntityClassificationFeatures.
            multi_label (bool, optional): [description]. Defaults to True.
            negative_label_id (int, optional): [description]. Defaults to 0.

        Raises:
            IncorrectHypothesisTemplateError: [description]
        """

        if not templates:
            if "{label}" not in hypothesis_template:
                raise IncorrectHypothesisTemplateError(
                    "The hypothesis_template argument must contain the '{label}' placeholder."
                )

            templates = {
                label: [hypothesis_template.replace("{label}", label)]
                for i, label in enumerate(labels)
                if i != negative_label_id
            }

        super().__init__(
            *args,
            name=name,
            required_variables=required_variables,
            additional_variables=additional_variables,
            labels=labels,
            templates=templates,
            valid_conditions=valid_conditions,
            features_class=features_class,
            multi_label=multi_label,
            negative_label_id=negative_label_id,
            **kwargs
        )



import unittest

from a2t.base import EntailmentClassifier
from a2t.tasks.text_classification import TopicClassificationFeatures, TopicClassificationTask

import os

os.environ["TOKENIZERS_PARALLELISM"] = "false"


class TestEntailmentModel(unittest.TestCase):
    def test_load_model(self):

        # Create a task
        task = TopicClassificationTask(
            "DummyTopic task", labels=["politics", "culture", "economy", "biology", "legal", "medicine", "business"]
        )

        features = [
            TopicClassificationFeatures(
                context="hospital: a health facility where patients receive treatment.", label="medicine"
            )
        ]

        nlp = EntailmentClassifier(use_tqdm=False)

        preds = nlp(task=task, features=features, negative_threshold=0.0, return_confidences=True, return_labels=True, topk=3)

        self.assertEqual(preds, [[("medicine", 0.8547821), ("biology", 0.036895804), ("business", 0.032091234)]])



from .mnli import (
    NLIRelationClassifierWithMappingHead,
    NLIRelationClassifier,
    REInputFeatures,
)
from .tacred import TACRED_LABELS, TACREDClassifier

__all__ = [
    "REInputFeatures",
    "NLIRelationClassifier",
    "NLIRelationClassifierWithMappingHead",
    "TACREDClassifier",
    "TACRED_LABELS",
]



from typing import Dict, List
from dataclasses import dataclass

from .base import BinaryTask, Features


@dataclass
class RelationClassificationFeatures(Features):
    X: str = None
    Y: str = None


class RelationClassificationTask(BinaryTask):
    """A class handler for Relation Classification task. It inherits from `BinaryTask` class.

    TODO: Add documentation.

    """

    def __init__(
        self,
        name: str,
        labels: List[str],
        *args,
        required_variables: List[str] = ["X", "Y"],
        additional_variables: List[str] = ["inst_type"],
        templates: Dict[str, List[str]] = None,
        valid_conditions: Dict[str, List[str]] = None,
        features_class: type = RelationClassificationFeatures,
        multi_label: bool = True,
        negative_label_id: int = 0,
        **kwargs
    ) -> None:
        """Initialization of a RelationClassificationTask task.

        Args:
            name (str): [description]
            labels (List[str]): [description]
            required_variables (List[str], optional): [description]. Defaults to ["X", "Y"].
            additional_variables (List[str], optional): [description]. Defaults to ["inst_type"].
            templates (Dict[str, List[str]], optional): [description]. Defaults to None.
            valid_conditions (Dict[str, List[str]], optional): [description]. Defaults to None.
            features_class (type, optional): [description]. Defaults to RelationClassificationFeatures.
            multi_label (bool, optional): [description]. Defaults to True.
            negative_label_id (int, optional): [description]. Defaults to 0.
        """
        super().__init__(
            *args,
            name=name,
            required_variables=required_variables,
            additional_variables=additional_variables,
            labels=labels,
            templates=templates,
            valid_conditions=valid_conditions,
            features_class=features_class,
            multi_label=multi_label,
            negative_label_id=negative_label_id,
            **kwargs
        )


@dataclass
class TACREDFeatures(Features):
    subj: str = None
    obj: str = None


class TACREDRelationClassificationTask(RelationClassificationTask):
    def __init__(
        self, labels: List[str], templates: Dict[str, List[str]], valid_conditions: Dict[str, List[str]], **kwargs
    ) -> None:
        for key in ["name", "required_variables", "additional_variables", "features_class", "multi_label", "negative_label_id"]:
            kwargs.pop(key, None)
        super().__init__(
            "TACRED Relation Classification task",
            labels=labels,
            required_variables=["subj", "obj"],
            additional_variables=["inst_type"],
            templates=templates,
            valid_conditions=valid_conditions,
            features_class=TACREDFeatures,
            multi_label=True,
            negative_label_id=0,
            **kwargs
        )



"""This repository contains the code for out of the box ready to use zero-shot classifiers among different tasks, 
such as Topic Labelling or Relation Extraction. It is built on top of 🤗 HuggingFace [Transformers](https://github.com/huggingface/transformers) 
library, so you are free to choose among hundreds of models. You can either, use a dataset specific classifier or define one 
yourself with just labels descriptions or templates!

## Getting Started

The framework is organized to differentiate **three** main components: the **data**, **task** and **inference**. 
Let's define a Topic Classifier that will classify sentences into the following topics: Politics, Culture, Economy, 
Biology, Legal, Medicine and Business.

### Defining the dataset

We will create a dummy dataset with only one instance to test our model. The `Dataset` object is intended to be 
used to load some data from a file and create the task features.

```python
from a2t.data import Dataset
from a2t.tasks import TopicClassificationFeatures

labels = [
    'politics', 'culture', 'economy', 'biology', 'legal', 'medicine', 'business'
]

class DummyTopicClassificationDataset(Dataset):
    def __init__(self) -> None:
        super().__init__(labels=labels)

        self.append(
            TopicClassificationFeatures(
                context="hospital: a health facility where patients"
                        " receive treatment.",
                label="medicine"
            )
        )

dataset = DummyTopicClassificationDataset()
```

You do not actually need to define a dataset, just a list of `Features` is enough.

### Defining the Task

The `Task` object will contain the **label verbalizations** and other task specific information. In this case it will be just the 
labels that we defined before. For more complex tasks like Relation Extraction you will probably need to define a set of `templates` 
and `valid_conditions`. This object should contain all the information regarding the task like the schema or ontology.

```python
from a2t.tasks import TopicClassificationTask

task = TopicClassificationTask(name="DummyTopic task", labels=labels)
```

### Defining the inference

The `EntailmentClassifier` object should be instantiated with the information about the pre-trained model and device information. You can 
make use of any entailment model available on 🤗 [Transformers](https://github.com/huggingface/transformers) that was
trained on some NLI dataset.

```python
from a2t.base import EntailmentClassifier

nlp = EntailmentClassifier(
    'roberta-large-mnli', 
    use_tqdm=False, 
    use_cuda=True, 
    half=True
)
```

### Putting all together

The following code is enough to run the model:

```python
predictions = nlp(
    task=task, 
    features=dataset, 
    return_labels=True, 
    return_confidences=True, 
    topk=3
)

print(predictions)
```
The result should be something close to this:
```python
[
    [('medicine', 0.8545), ('biology', 0.03693), ('business', 0.0322)]
]
```

## Information Extraction with Entailment

On the previous example we already saw how to creat a Topic Classifier that will classify the whole given text into a set of topic labels.
That kind of tasks are known as **Text Classification** tasks. On Information Extraction (IE) instead, we usually find tasks that require 
to classify spans inside a sentence (**Span Classification** tasks like NER) or relations between spans inside a sentence/document (**Tuple 
Classification tasks** like Relation Extraction). This framework differentiates the task types by the number of spans involved: if no spans 
are involved are `ZeroaryTask`, if 1 span is involved are `UnaryTask` and if 2 spans are involved the tasks are `BinaryTask`.

Let's build an small Relation Classifier based on [Sainz et al. (2021)](https://aclanthology.org/2021.emnlp-main.92/):

### Defining the templates for the task

We are going to build a small classifier that will classify entity pairs into the next relations:

* `per:city_of_death`: The `X` entity died in `Y` and `Y` is a city.
* `org:founded_by`: The `X` organization was founded by `Y` person.
* `no_relation`: No relation (among the predefined relations) exists between `X` and `Y`.

```python
labels = ["no_relation", "per:city_of_death", "org:founded_by"]
```

Then, we need to define the set of templates that verbalize each relation. These templates must contain **placeholders** for
two entity spans. In this case, we are going to use `X` and `Y` as they are already defined on `BinaryFeatures`.

```python
templates = {
    "per:city_of_death": [
        "{X} died in {Y}"
    ],
    "org:founded_by": [
        "{X} was founded by {Y}",
        "{Y} founded {X}"
    ]
}
```
We can also define a set of constraints to each of the labels to simplify the templates. For example, on [TACRED](https://nlp.stanford.edu/projects/tacred/)
exists both `per:city_of_death` and `per:country_of_death`, we could use the template `"{X} died in {Y}"` for both and differentiate them adding the 
`"PERSON:CITY"` constraint to `per:city_of_death` as follows:

```python
valid_conditions = {
    "per:city_of_death": [
        "PERSON:CITY",
        "PERSON:LOCATION"
    ],
    "org:founded_by": [
        "ORGANIZATION:PERSON"
    ]
}
```

Once we defined our **labels**, **templates** and **constraints** we can define our task as follows:

```python
from a2t.tasks import BinaryTask, BinaryFeatures

task = BinaryTask(
    name="Relation Classification task",
    required_variables=["X", "Y"],
    additional_variables=["inst_type"],
    labels=labels,
    templates=templates,
    valid_conditions=valid_conditions,
    negative_label_id=0,
    multi_label=True,
    features_class=BinaryFeatures
)
```

### Testing the Relation Classifier

At this point we have all we need to perform inferences on this task, let's see how it actually works:

```python
from a2t.base import EntailmentClassifier

nlp = EntailmentClassifier(
    "microsoft/deberta-v2-xlarge-mnli",
    use_tqdm=False,
    use_cuda=True, 
    half=True
)

test_examples = [
    BinaryFeatures(X='Billy Mays', Y='Tampa', inst_type='PERSON:CITY', context='Billy Mays, the bearded, boisterous pitchman who, as the undisputed king of TV yell and sell, became an unlikely pop culture icon, died at his home in Tampa, Fla, on Sunday', label='per:city_of_death'),
    BinaryFeatures(X='Old Lane Partners', Y='Pandit', inst_type='ORGANIZATION:PERSON', context='Pandit worked at the brokerage Morgan Stanley for about 11 years until 2005, when he and some Morgan Stanley colleagues quit and later founded the hedge fund Old Lane Partners.', label='org:founded_by'),
    BinaryFeatures(X='He', Y='University of Maryland in College Park', inst_type='PERSON:ORGANIZATION', context='He received an undergraduate degree from Morgan State University in 1950 and applied for admission to graduate school at the University of Maryland in College Park.', label='no_relation')
]

nlp(task=task, features=test_examples, return_labels=True, return_confidences=True)
```

The output should look like:

```python
[('per:city_of_death', 0.98828125),
 ('org:founded_by', 0.955078125),
 ('no_relation', 1.0)]
```

For more information consider reading the [Tasks](tasks/index.html) documentation. 
"""

__version__ = "0.3.0"

__pdoc__ = {
    "legacy": False,
    "evaluation": True,
    "tests": False,
    "tasks.base": False,
    "tasks.span_classification": False,
    "tasks.text_classification": False,
    "tasks.tuple_classification": False,
    "base.np_softmax": False,
    "base.np_sigmoid": False,
    "data.base": False,
    "data.tacred": False,
    "utils": False,
    "base.Classifier": False,
}



"""The module `data` implements different dataloaders or `Dataset`s for predefined tasks.
"""
from .tacred import TACREDRelationClassificationDataset
from .babeldomains import BabelDomainsTopicClassificationDataset
from .base import Dataset

PREDEFINED_DATASETS = {"tacred": TACREDRelationClassificationDataset, "babeldomains": BabelDomainsTopicClassificationDataset}

__all__ = ["Dataset", "TACREDRelationClassificationDataset", "BabelDomainsTopicClassificationDataset"]



import argparse
import json
import os
from pprint import pprint
from collections import Counter
import torch

import numpy as np
from sklearn.metrics import precision_recall_fscore_support, confusion_matrix

from .mnli import NLIRelationClassifierWithMappingHead, REInputFeatures
from .tacred import *
from .utils import find_optimal_threshold, apply_threshold

CLASSIFIERS = {"mnli-mapping": NLIRelationClassifierWithMappingHead}


def top_k_accuracy(output, labels, k=5):
    preds = np.argsort(output)[:, ::-1][:, :k]
    return sum(l in p and l > 0 for l, p in zip(labels, preds)) / (labels > 0).sum()


parser = argparse.ArgumentParser(
    prog="run_evaluation",
    description="Run a evaluation for each configuration.",
)
parser.add_argument(
    "input_file",
    type=str,
    default="data/tacred/dev.json",
    help="Dataset file.",
)
parser.add_argument(
    "--config",
    type=str,
    dest="config",
    help="Configuration file for the experiment.",
)
parser.add_argument("--basic", action="store_true", default=False)

args = parser.parse_args()

labels2id = (
    {label: i for i, label in enumerate(TACRED_LABELS)}
    if not args.basic
    else {label: i for i, label in enumerate(TACRED_BASIC_LABELS)}
)

with open(args.input_file, "rt") as f:
    features, labels = [], []
    for i, line in enumerate(json.load(f)):
        line["relation"] = (
            line["relation"] if not args.basic else TACRED_BASIC_LABELS_MAPPING.get(line["relation"], line["relation"])
        )
        features.append(
            REInputFeatures(
                subj=" ".join(line["token"][line["subj_start"] : line["subj_end"] + 1])
                .replace("-LRB-", "(")
                .replace("-RRB-", ")")
                .replace("-LSB-", "[")
                .replace("-RSB-", "]"),
                obj=" ".join(line["token"][line["obj_start"] : line["obj_end"] + 1])
                .replace("-LRB-", "(")
                .replace("-RRB-", ")")
                .replace("-LSB-", "[")
                .replace("-RSB-", "]"),
                pair_type=f"{line['subj_type']}:{line['obj_type']}",
                context=" ".join(line["token"])
                .replace("-LRB-", "(")
                .replace("-RRB-", ")")
                .replace("-LSB-", "[")
                .replace("-RSB-", "]"),
                label=line["relation"],
            )
        )
        labels.append(labels2id[line["relation"]])

labels = np.array(labels)

with open(args.config, "rt") as f:
    config = json.load(f)

LABEL_LIST = TACRED_BASIC_LABELS if args.basic else TACRED_LABELS

for configuration in config:
    n_labels = len(LABEL_LIST)
    os.makedirs(f"experiments/{configuration['name']}", exist_ok=True)
    _ = configuration.pop("negative_threshold", None)
    classifier = CLASSIFIERS[configuration["classification_model"]](negative_threshold=0.0, **configuration)
    output = classifier(
        features,
        batch_size=configuration["batch_size"],
        multiclass=configuration["multiclass"],
    )
    if not "use_threshold" in configuration or configuration["use_threshold"]:
        optimal_threshold, _ = find_optimal_threshold(labels, output)
        output_ = apply_threshold(output, threshold=optimal_threshold)
    else:
        output_ = output.argmax(-1)
    pre, rec, f1, _ = precision_recall_fscore_support(labels, output_, average="micro", labels=list(range(1, n_labels)))

    np.save(f"experiments/{configuration['name']}/output.npy", output)
    np.save(f"experiments/{configuration['name']}/labels.npy", labels)
    configuration["precision"] = pre
    configuration["recall"] = rec
    configuration["f1-score"] = f1
    configuration["top-1"] = top_k_accuracy(output, labels, k=1)
    configuration["top-3"] = top_k_accuracy(output, labels, k=3)
    configuration["top-5"] = top_k_accuracy(output, labels, k=5)
    configuration["topk-curve"] = [top_k_accuracy(output, labels, k=i) for i in range(1, n_labels + 1)]
    _ = configuration.pop("topk-curve", None)
    configuration["negative_threshold"] = optimal_threshold

    pprint(configuration)
    del classifier
    torch.cuda.empty_cache()


with open(args.config, "wt") as f:
    json.dump(config, f, indent=4)



from typing import List

from a2t.tasks.text_classification import TopicClassificationFeatures
from .base import Dataset


class BabelDomainsTopicClassificationDataset(Dataset):
    """A class to handle BabelDomains datasets.

    This class converts BabelDomains data files into a list of `TopicClassificationFeatures`.
    """

    def __init__(self, input_path: str, labels: List[str], *args, **kwargs) -> None:
        """
        Args:
            input_path (str): The path to the input file.
            labels (List[str]): The possible label set of the dataset.
        """
        super().__init__(labels=labels, *args, **kwargs)

        with open(input_path, "rt") as f:
            for line in f:
                _, label, context = line.strip().split("\t")
                self.append(TopicClassificationFeatures(context=context, label=label))



from pprint import pprint
from typing import Callable, List
import numpy as np
from sklearn.metrics import f1_score, precision_recall_fscore_support


def f1_score_(labels, preds, n_labels=None):
    n_labels = max(labels) + 1 if n_labels is None else n_labels
    return f1_score(labels, preds, labels=list(range(1, n_labels)), average="micro")


def individual_f1_score_(labels, preds, n_labels=None):
    n_labels = max(labels) + 1 if n_labels is None else n_labels
    return f1_score(labels, preds, labels=list(range(0, n_labels)), average=None)


def precision_recall_fscore_(labels, preds, n_labels: int = None):
    n_labels = max(labels) + 1 if n_labels is None else n_labels
    p, r, f, _ = precision_recall_fscore_support(labels, preds, labels=list(range(1, n_labels)), average="micro")
    return p, r, f


def apply_threshold(output, threshold=0.0, ignore_negative_prediction=True, negative_label_id: int = 0):
    """Applies a threshold to determine whether is a relation or not"""
    output_ = output.copy()
    if ignore_negative_prediction:
        output_[:, negative_label_id] = 0.0
    activations = (output_ >= threshold).sum(-1).astype(np.int)
    output_[activations == 0, negative_label_id] = 1.00

    return output_.argmax(-1)


def apply_individual_threshold(output, thresholds, ignore_negative_prediction=True):
    output_ = output.copy()
    if ignore_negative_prediction:
        output_[:, 0] = 0.0
    for i, threshold in enumerate(thresholds):
        if not i:
            continue
        activations = output_[:, i] < threshold
        output_[activations, i] = 0.0

    return output_.argmax(-1)


def find_optimal_threshold(
    labels,
    output,
    granularity=1000,
    metric=f1_score_,
    n_labels=None,
    negative_label_id: int = 0,
    apply_threshold_fn: Callable = apply_threshold,
):
    thresholds = np.linspace(0, 1, granularity)
    values = []
    for t in thresholds:
        preds = apply_threshold_fn(output, threshold=t, negative_label_id=negative_label_id)
        values.append(metric(labels, preds, n_labels=n_labels))

    best_metric_id = np.argmax(values)
    best_threshold = thresholds[best_metric_id]

    return best_threshold, values[best_metric_id]


def find_optimal_individual_threshold(
    labels,
    output,
    granularity=1000,
    indv_metric=individual_f1_score_,
    metric=f1_score_,
    n_labels=None,
    default=0.9,
):
    thresholds = np.linspace(0, 1, granularity)
    values = []
    for t in thresholds:
        preds = apply_individual_threshold(output, thresholds=[t] * output.shape[1])
        values.append(indv_metric(labels, preds, n_labels=n_labels))

    best_metric_id = np.argmax(values, 0)
    best_threshold = thresholds[best_metric_id]

    # Fill the thresholds of unseen arguments with default=.5
    if n_labels is not None:
        for i in range(n_labels):
            if i not in np.unique(labels):
                best_threshold[i] = default

    return best_threshold, metric(labels, apply_individual_threshold(output, thresholds=best_threshold))


def apply_multi_label_threshold(output, threshold, **kwargs):
    return [[i for i, coef in enumerate(row) if coef >= threshold] for row in output]


def multi_label_metrics(y_true: List[int], y_pred: List[int], **kwargs):
    tp, total_t, total_p = 0, 0, 0
    for y_t, y_p in zip(y_true, y_pred):
        tp += sum([p in y_t for p in y_p])
        total_t += len(y_t)
        total_p += len(y_p)

    precision = tp / total_p if total_p > 0 else 0.0
    recall = tp / total_t if total_t > 0 else 0.0
    f1 = 2 * (precision * recall) / (precision + recall) if precision + recall > 0 else 0.0

    return precision, recall, f1


if __name__ == "__main__":
    y_true = [[0], [1, 2], [3]]
    output = np.array([[0.9, 0.4, 0.2], [0.7, 0.8, 0.3], [0.2, 0.6, 0.9]])
    threshold, _ = find_optimal_threshold(
        y_true,
        output,
        metric=lambda y_true, y_pred, **kwargs: multi_label_metrics(y_true, y_pred, **kwargs)[-1],
        apply_threshold_fn=apply_multi_label_threshold,
    )
    y_pred = apply_multi_label_threshold(output, threshold)
    pre, rec, f1 = multi_label_metrics(y_true, y_pred)
    pprint(y_pred)
    pprint(threshold)
    print(pre, rec, f1)



from .mlm import MLMTopicClassifier
from .mnli import NLITopicClassifierWithMappingHead, NLITopicClassifier
from .nsp import NSPTopicClassifier
from .babeldomains import BabelDomainsClassifier
from .wndomains import WNDomainsClassifier

__all__ = [
    "NLITopicClassifierWithMappingHead",
    "NLITopicClassifier",
    "MLMTopicClassifier",
    "NSPTopicClassifier",
    "BabelDomainsClassifier",
    "WNDomainsClassifier",
]



import setuptools
import a2t

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="a2t",
    version=a2t.__version__,
    author="Oscar Sainz",
    author_email="oscar.sainz@ehu.eus",
    description="Ask2Transformers is a library for zero-shot classification based on Transformers.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/osainz59/Ask2Transformers",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["transformers", "tqdm", "torch", "numpy", "scikit-learn"],
)



import unittest
import tempfile
from dataclasses import dataclass
from a2t.base import EntailmentClassifier

from a2t.tasks.base import IncorrectFeatureTypeError, Task, Features, ZeroaryTask, UnaryTask, BinaryTask
from a2t.tasks.span_classification import NamedEntityClassificationFeatures, NamedEntityClassificationTask
from a2t.tasks.text_classification import (
    IncorrectHypothesisTemplateError,
    TopicClassificationTask,
    TopicClassificationFeatures,
)

import os

os.environ["TOKENIZERS_PARALLELISM"] = "false"


class TestTasks(unittest.TestCase):
    def test_assert_minimal_constraints(self):

        # Check number of labels greater than 0
        self.assertRaises(AssertionError, Task)

        # Check templates keys
        self.assertRaises(AssertionError, Task, labels=["O"], templates={"B": []})

        # Check valid conditions keys
        self.assertRaises(AssertionError, Task, labels=["O"], templates={"O": []}, valid_conditions={"B": []})

        # Check negative label id to be smaller than the number of labels
        self.assertRaises(AssertionError, Task, labels=["O"], templates={"O": []}, negative_label_id=1)

        # Check required and additional variables
        @dataclass
        class DummyFeatures(Features):
            X: str = None

        class DummyTask(Task):
            def _assert_constraints(self):
                pass

        _ = DummyTask(required_variables=["X"], labels=["O"], templates={"O": []}, features_class=DummyFeatures)

        self.assertRaises(AssertionError, DummyTask, required_variables=["X"], labels=["O"], templates={"O": []})

        _ = DummyTask(
            required_variables=["X"], labels=["O"], templates={"O": ["{X} is a Potato."]}, features_class=DummyFeatures
        )

        self.assertRaises(
            AssertionError, DummyTask, required_variables=["X"], labels=["O"], templates={"O": ["{Y} is a Potato."]}
        )

        # Abstract class should not be instantiated
        self.assertRaises(NotImplementedError, Task, labels=["O"])

    def test_assert_constraints(self):
        _ = ZeroaryTask(labels=["O"])

        @dataclass
        class UnaryFeatures(Features):
            X: str = None

        _ = UnaryTask(labels=["O"], required_variables=["X"], features_class=UnaryFeatures)

        @dataclass
        class BinaryFeatures(Features):
            X: str = None
            Y: str = None

        _ = BinaryTask(labels=["O"], required_variables=["X", "Y"], features_class=BinaryFeatures)

    def test_assert_feature_type(self):
        @dataclass
        class UnaryFeatures(Features):
            X: str = None

        unary_task = UnaryTask(labels=["O"], required_variables=["X"], features_class=UnaryFeatures)

        unary_features = [UnaryFeatures(context="", label="", X=str(i)) for i in range(10)]

        @dataclass
        class BinaryFeatures(Features):
            X: str = None
            Y: str = None

        binary_task = BinaryTask(labels=["O"], required_variables=["X", "Y"], features_class=BinaryFeatures)

        binary_features = [BinaryFeatures(context="", label="", X=str(i), Y=str(i + 10)) for i in range(10)]

        # Correct
        unary_task.assert_features_class(unary_features)
        binary_task.assert_features_class(binary_features)

        # Incorrect
        self.assertRaises(IncorrectFeatureTypeError, unary_task.assert_features_class, binary_features)
        self.assertRaises(IncorrectFeatureTypeError, binary_task.assert_features_class, unary_features)

    def test_save_and_load(self):
        task = ZeroaryTask(
            "Topic Classification task.",
            labels=["Medicine", "Sports", "Politics"],
            templates={
                "Medicine": ["Topic about health.", "Topic about medicine."],
                "Sports": ["Topic about footbal.", "Topic about sports."],
                "Politics": ["Topic about corruption.", "Topic about politics."],
            },
        )

        with tempfile.TemporaryDirectory() as tmpdir:
            task.to_config(os.path.join(tmpdir, "task.config.json"))
            task2 = ZeroaryTask.from_config(os.path.join(tmpdir, "task.config.json"))

        self.assertEqual(task, task2)


class TestTopicClassification(unittest.TestCase):
    def test_hypothesis_generation(self):
        labels = ["Culture", "Health", "Army"]

        features = [
            TopicClassificationFeatures(context="hospital: a health facility where patients receive treatment.", label="Health")
        ]

        hypothesis_template = "The domain of the sentence is {label}."

        task = TopicClassificationTask("Test Topic task", labels=labels, hypothesis_template=hypothesis_template)

        premise_hypothesis_pairs = task.generate_premise_hypotheses_pairs(features)
        hypotheses = [hypothesis_template.format(label=label) for label in labels]

        self.assertTrue(len(premise_hypothesis_pairs) == len(hypotheses), "The amount of hypothesis generated is not correct.")

        self.assertTrue(
            all(correct in generated for correct, generated in zip(hypotheses, premise_hypothesis_pairs)),
            "The generated hypothesis does not match the correct ones.",
        )

        self.assertRaises(
            IncorrectHypothesisTemplateError,
            TopicClassificationTask,
            "Test Topic task",
            labels=labels,
            hypothesis_template="This template does not contain a label placeholder.",
        )


class TestNamedEntityClassificationTask(unittest.TestCase):
    def test_hypothesis_generation(self):
        labels = ["O", "Person", "Location", "Organization"]

        valid_conditions = {"Person": ["*"], "Location": ["NNP"], "Organization": ["NNP"]}

        task = NamedEntityClassificationTask("Dummy NER task", labels=labels, valid_conditions=valid_conditions)

        features = [
            NamedEntityClassificationFeatures(
                context="Peter won an award in London.", label="Person", inst_type="NNP", X="Peter"
            ),
            NamedEntityClassificationFeatures(
                context="Peter won an award in London.", label="Location", inst_type="NNP", X="London"
            ),
        ]

        self.assertEqual(
            task.generate_premise_hypotheses_pairs(features),
            [
                "Peter won an award in London. </s> Peter is a Person.",
                "Peter won an award in London. </s> Peter is a Location.",
                "Peter won an award in London. </s> Peter is a Organization.",
                "Peter won an award in London. </s> London is a Person.",
                "Peter won an award in London. </s> London is a Location.",
                "Peter won an award in London. </s> London is a Organization.",
            ],
        )

    def test_inference(self):
        labels = ["O", "Person", "Location", "Organization"]

        valid_conditions = {"Person": ["NNP"], "Location": ["NNP"], "Organization": ["NNP"]}

        task = NamedEntityClassificationTask("Dummy NER task", labels=labels, valid_conditions=valid_conditions)

        features = [
            NamedEntityClassificationFeatures(
                context="Peter won an award in London.", label="Person", inst_type="NNP", X="Peter"
            ),
            NamedEntityClassificationFeatures(
                context="Peter won an award in London.", label="Location", inst_type="NNP", X="London"
            ),
        ]

        nlp = EntailmentClassifier(use_cuda=False, use_tqdm=False)

        preds = nlp(task=task, features=features, negative_threshold=0.5, return_labels=True, topk=1)

        self.assertEqual(preds, [feature.label for feature in features])



from typing import List
import json

from a2t.tasks.tuple_classification import TACREDFeatures
from .base import Dataset


class TACREDRelationClassificationDataset(Dataset):
    """A class to handle TACRED datasets.

    This class converts TACRED data files into a list of `TACREDFeatures`.
    """

    def __init__(self, input_path: str, labels: List[str], *args, **kwargs) -> None:
        """
        Args:
            input_path (str): The path to the input file.
            labels (List[str]): The possible label set of the dataset.
        """
        super().__init__(labels=labels, *args, **kwargs)

        with open(input_path, "rt") as f:
            for i, line in enumerate(json.load(f)):
                self.append(
                    TACREDFeatures(
                        subj=" ".join(line["token"][line["subj_start"] : line["subj_end"] + 1])
                        .replace("-LRB-", "(")
                        .replace("-RRB-", ")")
                        .replace("-LSB-", "[")
                        .replace("-RSB-", "]"),
                        obj=" ".join(line["token"][line["obj_start"] : line["obj_end"] + 1])
                        .replace("-LRB-", "(")
                        .replace("-RRB-", ")")
                        .replace("-LSB-", "[")
                        .replace("-RSB-", "]"),
                        inst_type=f"{line['subj_type']}:{line['obj_type']}",
                        context=" ".join(line["token"])
                        .replace("-LRB-", "(")
                        .replace("-RRB-", ")")
                        .replace("-LSB-", "[")
                        .replace("-RSB-", "]"),
                        label=line["relation"],
                    )
                )



from typing import List

import numpy as np

from . import NLITopicClassifierWithMappingHead

WNDOMAINS_TOPICS = [
    "play",
    "tourism",
    "politics",
    "alimentation",
    "astrology",
    "biology",
    "history",
    "physics",
    "pedagogy",
    "color",
    "medicine",
    "psychology",
    "commerce",
    "fashion",
    "anthropology",
    "astronomy",
    "mathematics",
    "administration",
    "agriculture",
    "computer science",
    "earth",
    "law",
    "number",
    "archaeology",
    "literature",
    "time period",
    "quality",
    "engineering",
    "architecture",
    "veterinary",
    "military",
    "transport",
    "body care",
    "metrology",
    "industry",
    "chemistry",
    "telecommunication",
    "art",
    "sexuality",
    "sport",
    "religion",
    "person",
    "philosophy",
    "linguistics",
    "publishing",
    "artisanship",
    "sociology",
    "economy",
]

WNDOMAINS_TOPIC_MAPPING = {
    "acoustics": "physics",
    "administration": "administration",
    "aeronautic": "transport",
    "agriculture": "agriculture",
    "alimentation": "alimentation",
    "anatomy": "biology",
    "anthropology": "anthropology",
    "archaeology": "archaeology",
    "archery": "sport",
    "architecture": "architecture",
    "art": "art",
    "artisanship": "artisanship",
    "astrology": "astrology",
    "astronautics": "engineering",
    "astronomy": "astronomy",
    "athletics": "sport",
    "atomic physic": "physics",
    "auto": "transport",
    "badminton": "sport",
    "banking": "economy",
    "baseball": "sport",
    "basketball": "sport",
    "betting": "play",
    "biochemistry": "biology",
    "biology": "biology",
    "body care": "body care",
    "book keeping": "economy",
    "botany": "biology",
    "bowling": "sport",
    "boxing": "sport",
    "building industry": "architecture",
    "card": "play",
    "chemistry": "chemistry",
    "chess": "play",
    "cinema": "telecommunication",
    "color": "color",
    "commerce": "commerce",
    "computer science": "computer science",
    "cricket": "sport",
    "cycling": "sport",
    "dance": "art",
    "dentistry": "medicine",
    "diplomacy": "politics",
    "diving": "sport",
    "drawing": "art",
    "earth": "earth",
    "ecology": "biology",
    "economy": "economy",
    "electricity": "physics",
    "electronics": "physics",
    "electrotechnics": "engineering",
    "engineering": "engineering",
    "enterprise": "economy",
    "entomology": "biology",
    "ethnology": "anthropology",
    "exchange": "economy",
    "fashion": "fashion",
    "fencing": "sport",
    "fishing": "sport",
    "folklore": "anthropology",
    "football": "sport",
    "furniture": "architecture",
    "gas": "physics",
    "gastronomy": "alimentation",
    "genetics": "biology",
    "geography": "earth",
    "geology": "earth",
    "geometry": "mathematics",
    "golf": "sport",
    "grammar": "linguistics",
    "heraldry": "history",
    "history": "history",
    "hockey": "sport",
    "hunting": "sport",
    "hydraulics": "engineering",
    "industry": "industry",
    "insurance": "economy",
    "jewellery": "art",
    "law": "law",
    "linguistics": "linguistics",
    "literature": "literature",
    "mathematics": "mathematics",
    "mechanics": "engineering",
    "medicine": "medicine",
    "merchant navy": "transport",
    "meteorology": "earth",
    "metrology": "metrology",
    "military": "military",
    "money": "economy",
    "mountaineering": "sport",
    "music": "art",
    "mythology": "religion",
    "number": "number",
    "numismatics": "art",
    "occultism": "religion",
    "oceanography": "earth",
    "optics": "physics",
    "painting": "art",
    "paleontology": "earth",
    "pedagogy": "pedagogy",
    "person": "person",
    "pharmacy": "medicine",
    "philately": "art",
    "philology": "literature",
    "philosophy": "philosophy",
    "photography": "art",
    "physics": "physics",
    "physiology": "biology",
    "plastic arts": "art",
    "play": "play",
    "politics": "politics",
    "post": "telecommunication",
    "psychiatry": "medicine",
    "psychoanalysis": "psychology",
    "psychology": "psychology",
    "publishing": "publishing",
    "quality": "quality",
    "racing": "sport",
    "radio": "telecommunication",
    "radiology": "medicine",
    "railway": "transport",
    "religion": "religion",
    "roman catholic": "religion",
    "rowing": "sport",
    "rugby": "sport",
    "school": "pedagogy",
    "sculpture": "art",
    "sexuality": "sexuality",
    "skating": "sport",
    "skiing": "sport",
    "soccer": "sport",
    "sociology": "sociology",
    "sport": "sport",
    "statistics": "mathematics",
    "sub": "sport",
    "surgery": "medicine",
    "swimming": "sport",
    "table tennis": "sport",
    "tax": "economy",
    "telecommunication": "telecommunication",
    "telegraphy": "telecommunication",
    "telephony": "telecommunication",
    "tennis": "sport",
    "textiles": "industry",
    "theatre": "art",
    "theology": "religion",
    "time period": "time period",
    "topography": "astronomy",
    "tourism": "tourism",
    "town planning": "architecture",
    "transport": "transport",
    "tv": "telecommunication",
    "university": "pedagogy",
    "veterinary": "veterinary",
    "volleyball": "sport",
    "wrestling": "sport",
    "zoology": "biology",
    "zootechnics": "veterinary",
}


class WNDomainsClassifier(NLITopicClassifierWithMappingHead):
    """WNDomainsClassifier

    Specific class for topic classification using WNDomains topic set.
    """

    def __init__(self, **kwargs):
        super(WNDomainsClassifier, self).__init__(
            pretrained_model="roberta-large-mnli",
            labels=WNDOMAINS_TOPICS,
            topic_mapping=WNDOMAINS_TOPIC_MAPPING,
            query_phrase="The domain of the sentence is about",
            entailment_position=2,
            **kwargs
        )



"""The module `base` implements all the basic methods to perform the inference, including the `EntailmentClassifier`.
"""
import os
import sys
import gc
from typing import List

import numpy as np
import torch

try:
    from tqdm import tqdm

    _use_tqdm = True
except ImportError:
    _use_tqdm = False

from transformers import AutoConfig, AutoTokenizer, AutoModelForSequenceClassification

from .tasks import Features, Task

try:
    import transformers

    transformers.logging.set_verbosity_error()
except ImportError:
    pass


def np_softmax(x, dim=-1):
    e = np.exp(x)
    return e / np.sum(e, axis=dim, keepdims=True)


def np_sigmoid(x, dim=-1):
    return 1 / (1 + np.exp(-x))


class Classifier(object):
    """Abstact classifier class."""

    def __init__(
        self, labels: List[str], pretrained_model: str = "roberta-large-mnli", use_cuda=True, half=False, verbose=True
    ):
        super().__init__()

        self.device = torch.device("cuda" if torch.cuda.is_available() and use_cuda else "cpu")
        self.labels = labels
        self.use_cuda = use_cuda
        self.half = half
        self.verbose = verbose

        # Supress stdout printing for model downloads
        if not verbose:
            sys.stdout = open(os.devnull, "w")
            self._initialize(pretrained_model)
            sys.stdout = sys.__stdout__
        else:
            self._initialize(pretrained_model)

        self.model = self.model.to(self.device)
        self.model = self.model.eval()
        if self.use_cuda and self.half and torch.cuda.is_available():
            self.model = self.model.half()

    def _initialize(self, pretrained_model):
        raise NotImplementedError

    def __call__(self, context, batch_size=1):
        raise NotImplementedError

    def clear_gpu_memory(self):
        self.model.cpu()
        del self.model
        gc.collect()
        torch.cuda.empty_cache()


class EntailmentClassifier(Classifier):
    """General purpose Entailment based classifier.

    This class contains the code for entailment-based zero-shot classification inference. It is pretended to be
    task and data independent.
    """

    def __init__(
        self,
        pretrained_model: str = "roberta-large-mnli",
        use_cuda: bool = True,
        half: bool = False,
        verbose: bool = True,
        use_tqdm: bool = True,
        **kwargs
    ):
        """
        Args:
            pretrained_model (str, optional): The name or path of the pretrained model. Defaults to "roberta-large-mnli".
            use_cuda (bool, optional): Use the GPU if possible. Defaults to True.
            half (bool, optional): Use half precision if possible. Defaults to False.
            verbose (bool, optional): Output log information. Defaults to True.
        """
        super().__init__(None, pretrained_model, use_cuda, half, verbose)
        self.use_tqdm = use_tqdm and _use_tqdm

    def _initialize(self, pretrained_model: str):
        self.tokenizer = AutoTokenizer.from_pretrained(pretrained_model)
        self.model = AutoModelForSequenceClassification.from_pretrained(pretrained_model)
        self.config = AutoConfig.from_pretrained(pretrained_model)
        self.ent_pos = self.config.label2id.get("ENTAILMENT", self.config.label2id.get("entailment", None))
        if self.ent_pos is None:
            raise ValueError("The model config must contain ENTAILMENT label in the label2id dict.")
        else:
            self.ent_pos = int(self.ent_pos)

    @staticmethod
    def apply_threshold(
        output: np.ndarray,
        threshold: float = 0.0,
        ignore_negative_prediction: bool = True,
        application_type: str = "prediction",
    ) -> np.ndarray:
        """

        Args:
            output (ndarray): (batch_size, n_labels) The predicted probabilities.
            threshold (float): The threshold value to apply.
            ignore_negative_prediction (bool): Ignore the negative prediction probabilites. Default to True.
            application_type (str): How to apply the threshold: Options:

                * **"prediction"**: Set to 1.0 the probability of the negative class if the no prediction is higher than the threshold.
                * **"mask"**: Set to 0.0 the probabilities of the positive classes that are lower or equal to the threshold.
        """
        output_ = output.copy()
        if ignore_negative_prediction:
            output_[:, 0] = 0.0
        if application_type == "prediction":
            activations = (output_ >= threshold).sum(-1).astype(int)
            output_[activations == 0, 0] = 1.00
        elif application_type == "mask":
            activations = output_ < threshold
            output_[activations] = 0.0
        else:
            raise ValueError("""application_type argument must be "prediction" or "mask".""")

        return output_

    def __call__(
        self,
        task: Task,
        features: List[Features],
        negative_threshold: float = 0.5,
        topk: int = 1,
        return_labels: bool = False,
        return_confidences: bool = False,
        ignore_negative_prediction: bool = False,
        return_raw_output: bool = False,
        **kwargs
    ) -> List:
        """Call method for the EntailmentClassifier.


        TODO: Add output documentation.

        Args:
            task (Task): The task instance used for inference.
            features (List[Features]): The list of features to classify.
            negative_threshold (float, optional): The threshold used if necessary. Defaults to 0.5.
            topk (int, optional): Return the first `k` predictions with higher probabilities. Defaults to 1.
            return_labels (bool, optional): Whether to return the label ids or names. Defaults to False (ids).
            return_confidences (bool, optional): Whether to return prediction confidences or not. Defaults to False.
            ignore_negative_prediction (bool, optional): Whether to ignore the predictions of the negative class. Defaults to False.
            return_raw_output (bool, optional): Return the raw output along with the processed one. Defaults to False.

        Returns:
            List: A list with the predictions.
        """
        task.assert_features_class(features)

        outputs = []
        iterator = features if not self.use_tqdm else tqdm(features, total=len(features))
        with torch.no_grad():
            for feature in iterator:
                sentence_pairs = task.generate_premise_hypotheses_pairs([feature], self.tokenizer.sep_token)
                data = self.tokenizer(sentence_pairs, return_tensors="pt", padding=True).input_ids
                data = data.to(self.device)
                output = self.model(data)[0].detach().cpu().numpy()
                outputs.append(output)

        outputs = np.vstack(outputs)

        if task.multi_label:
            outputs = np.exp(outputs) / np.exp(outputs).sum(-1, keepdims=True)
        outputs = outputs[..., self.ent_pos].reshape(len(features), -1)

        preds = task.reverse_to_labels(outputs)
        if not task.multi_label:
            preds = np_softmax(preds)

        preds = task.apply_valid_conditions(features, preds)

        apply_threshold = task.multi_label and negative_threshold > 0
        if apply_threshold:
            preds = self.apply_threshold(
                preds, threshold=negative_threshold, ignore_negative_prediction=ignore_negative_prediction
            )

        predictions = np.argsort(preds, -1)[:, ::-1]
        if topk > 0:
            predictions = predictions[:, :topk]
        if return_labels:
            predictions = task.idx2label(predictions)
        if return_confidences:
            confidences = np.sort(preds, -1)[:, ::-1]
            if topk > 0:
                confidences = confidences[:, :topk]

            predictions = np.stack((predictions, confidences), -1).tolist()
            predictions = [
                [(int(label), float(conf)) if not return_labels else (label, float(conf)) for label, conf in row]
                for row in predictions
            ]
        else:
            predictions = predictions.tolist()
        if topk == 1:
            predictions = [row[0] for row in predictions]

        if return_raw_output:
            return (predictions, preds)
        else:
            return predictions


__pdoc__ = {"EntailmentClassifier.__call__": True}



"""Main evaluation script.

Please consider making a copy and customizing it yourself if you aim to use a custom class that is not already 
defined on the library.

### Usage

```bash 
a2t.evaluation [-h] [--config CONFIG]

optional arguments:
  -h, --help       show this help message and exit
  --config CONFIG  Config with task (schema) and data information.
```

### Configuration file

A configuration file containing the task and evaluation information should look like this:

```json
{
    "name": "BabelDomains",
    "task_name": "topic-classification",
    "features_class": "a2t.tasks.text_classification.TopicClassificationFeatures",
    "hypothesis_template": "The domain of the sentence is about {label}.",
    "nli_models": [
        "roberta-large-mnli"
    ],
    "labels": [
        "Animals",
        "Art, architecture, and archaeology",
        "Biology",
        "Business, economics, and finance",
        "Chemistry and mineralogy",
        "Computing",
        "Culture and society",
        ...
        "Royalty and nobility",
        "Sport and recreation",
        "Textile and clothing",
        "Transport and travel",
        "Warfare and defense"
    ],
    "preprocess_labels": true,
    "dataset": "babeldomains",
    "test_path": "data/babeldomains.domain.gloss.tsv",
    "use_cuda": true,
    "half": true
}
```

"""
import argparse
import json
import os
from pprint import pprint
from types import SimpleNamespace
import numpy as np
import torch

from a2t.tasks import PREDEFINED_TASKS
from a2t.data import PREDEFINED_DATASETS
from a2t.base import EntailmentClassifier


def main(args):

    with open(args.config, "rt") as f:
        config = SimpleNamespace(**json.load(f))

    os.makedirs(f"experiments/{config.name}", exist_ok=True)

    task_class, _ = PREDEFINED_TASKS[config.task_name]
    task = task_class.from_config(args.config)  # (**vars(config))

    dataset_class = PREDEFINED_DATASETS[config.dataset]

    assert hasattr(config, "dev_path") or hasattr(config, "test_path"), "At least a test or dev path must be provided."

    # Run dev evaluation
    if hasattr(config, "dev_path"):
        dev_dataset = dataset_class(config.dev_path, task.labels)
    else:
        dev_dataset = None

    if hasattr(config, "test_path"):
        test_dataset = dataset_class(config.test_path, task.labels)
    else:
        test_dataset = None

    results = {}
    for pretrained_model in config.nli_models:

        nlp = EntailmentClassifier(pretrained_model, **vars(config))

        results[pretrained_model] = {}

        if dev_dataset:
            _, output = nlp(task=task, features=dev_dataset, negative_threshold=0.0, return_raw_output=True, **vars(config))

            dev_labels = dev_dataset.labels

            # Save the output
            os.makedirs(
                f"experiments/{config.name}/{pretrained_model}/dev",
                exist_ok=True,
            )
            np.save(
                f"experiments/{config.name}/{pretrained_model}/dev/output.npy",
                output,
            )
            np.save(
                f"experiments/{config.name}/{pretrained_model}/dev/labels.npy",
                dev_labels,
            )

            # If dev data then optimize the threshold on it
            dev_results = task.compute_metrics(dev_labels, output, threshold="optimize")
            results[pretrained_model]["dev"] = dev_results
            pprint(results)

            with open(f"experiments/{config.name}/results.json", "wt") as f:
                json.dump(results, f, indent=4)

        if test_dataset:
            _, output = nlp(task=task, features=test_dataset, negative_threshold=0.0, return_raw_output=True, **vars(config))

            test_labels = test_dataset.labels

            # Save the output
            os.makedirs(
                f"experiments/{config.name}/{pretrained_model}/test",
                exist_ok=True,
            )
            np.save(
                f"experiments/{config.name}/{pretrained_model}/test/output.npy",
                output,
            )
            np.save(
                f"experiments/{config.name}/{pretrained_model}/test/labels.npy",
                test_labels,
            )

            optimal_threshold = 0.5 if not dev_dataset else dev_results["optimal_threshold"]
            test_results = task.compute_metrics(test_labels, output, threshold=optimal_threshold)
            results[pretrained_model]["test"] = test_results

            with open(f"experiments/{config.name}/results.json", "wt") as f:
                json.dump(results, f, indent=4)

        nlp.clear_gpu_memory()
        del nlp
        torch.cuda.empty_cache()


if __name__ == "__main__":
    parser = argparse.ArgumentParser("a2t.evaluation")
    parser.add_argument("--config", type=str, help="Config with task (schema) and data information.")

    args = parser.parse_args()
    main(args)



"""The module `tasks` contains the code related to the `Task` definition.

![Task taxonomy.](https://raw.githubusercontent.com/osainz59/Ask2Transformers/master/imgs/task_taxonomy.svg)

The tasks on this module are organized based on the number of spans to classify:

* `ZeroaryTask`: are tasks like Text Classification, where the aim is to classify the given text into a set of predefined
labels.
* `UnaryTask`: are tasks like Named Entity Classification, where the object to classify is an span within a text.
* `BinaryTask`: are tasks like Relation Classification, where what is actually classified is the relation between two spans in a text.

There are also more specific predefined task classes like `TopicClassificationTask` that includes helpful code and default values for 
the given task. You can either create create a task specific class or instantiate one of the predefined ones.

In addition to the `Task` class a `Features` class must be defined. The `Features` class will define which type of information is
used during the classification. For example, for a `UnaryTask` a `context` and some variable `X` for the span are needed. This class
will also be used to instantiate the task data instances.
"""

from .base import Task, ZeroaryTask, UnaryTask, BinaryTask, Features, ZeroaryFeatures, UnaryFeatures, BinaryFeatures
from .text_classification import TopicClassificationFeatures, TopicClassificationTask
from .span_classification import NamedEntityClassificationFeatures, NamedEntityClassificationTask
from .tuple_classification import (
    RelationClassificationFeatures,
    RelationClassificationTask,
    TACREDRelationClassificationTask,
    TACREDFeatures,
)


PREDEFINED_TASKS = {
    "zero-ary": (ZeroaryTask, ZeroaryFeatures),
    "unary": (UnaryTask, UnaryFeatures),
    "binary": (BinaryTask, BinaryFeatures),
    "topic-classification": (TopicClassificationTask, TopicClassificationFeatures),
    "named-entity-classification": (NamedEntityClassificationTask, NamedEntityClassificationFeatures),
    "relation-classification": (RelationClassificationTask, RelationClassificationFeatures),
    "tacred": (TACREDRelationClassificationTask, TACREDFeatures),
}


__all__ = [
    "Task",
    "Features",
    "ZeroaryTask",
    "ZeroaryFeatures",
    "UnaryTask",
    "UnaryFeatures",
    "BinaryTask",
    "BinaryFeatures",
    "TopicClassificationFeatures",
    "TopicClassificationTask",
    "TACREDFeatures",
    "TACREDRelationClassificationTask",
    "PREDEFINED_TASKS",
]

# Ignore __dataclass_fields__ variables on documentation
__pdoc__ = {
    **{
        f"{_class}.{varname}": False
        for _class in __all__
        if hasattr(eval(_class), "__dataclass_fields__")
        for varname in eval(_class).__dataclass_fields__.keys()
    }
}
